# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import os
from pathlib import Path
from shutil import which
from textwrap import indent, shorten

import click
from boltons.cacheutils import cachedproperty
from boltons.iterutils import flatten
from boltons.strutils import strip_ansi
from boltons.typeutils import classproperty

from . import logger
from .bitbar import run
from .platform import CURRENT_OS_ID
from .version import parse_version

# Rendering format of CLI in JSON fields.
CLI_FORMATS = frozenset(["plain", "fragments", "bitbar"])


class CLIError(Exception):

    """ An error occurred when running package manager CLI. """

    def __init__(self, code, output, error):
        """ The exception internally keeps the result of CLI execution. """
        super(CLIError, self).__init__()
        self.code = code
        self.output = output
        self.error = error

    def __str__(self):
        """ Human-readable error. """
        margin = " " * 2
        return indent(
            ("\nReturn code: {}\n" "Output:\n{}\n" "Error:\n{}").format(
                self.code,
                indent(str(self.output), margin),
                indent(str(self.error), margin),
            ),
            margin,
        )

    def __repr__(self):
        return "<{}({}, {!r})>".format(
            self.__class__.__name__,
            self.code,
            shorten(" ".join(self.error.split()), width=60, placeholder="(...)"),
        )


class PackageManager:

    """Base class from which all package manager definitions must inherits."""

    # List of platforms supported by the manager.
    platforms = frozenset()

    # Version requirement specifier.
    requirement = None

    # List of additional path to help mpm hunt down the package manager CLI.
    # Should be a list of strings whose order dictatate the search sequence.
    # Most of the time unnecessay: the `cli_path()` method works well on all
    # platforms.
    cli_search_path = []

    # Global list of options used for each call to the package manager CLI.
    # Might be of use to force silencing or high verbosity.
    global_args = []

    # Tell the manager either to raise or continue on errors.
    raise_on_cli_error = False

    # Some managers have the ability to report or ignore packages
    # possessing their own auto-update mecanism.
    ignore_auto_updates = True

    def __init__(self):
        # Log of all encountered CLI errors.
        self.cli_errors = []

    @classproperty
    def id(cls):
        """Return package manager's ID. Defaults based on class name.

        This ID must be unique among all package manager definitions and
        lower-case as they're used as feature flags for the :command:`mpm` CLI.
        """
        return cls.__name__.lower()

    @classproperty
    def name(cls):
        """Return package manager's common name. Defaults based on class name."""
        return cls.__name__

    @classproperty
    def cli_name(cls):
        """Package manager's CLI name.

        Is derived by default from the manager's ID.
        """
        return cls.id

    @classproperty
    def virtual(cls):
        """Should we expose the package manager to the user?

        Virtual package manager are just skeleton classes used to factorize
        code among managers of the same family.
        """
        return cls.__name__ == "PackageManager" or not cls.cli_name

    @cachedproperty
    def cli_path(self):
        """Fully qualified path to the package manager CLI.

        Automaticaly search the location of the CLI in the system.

        Returns `None` if CLI is not found or is not a file.
        """
        if not self.cli_name:
            return None
        env_path = ":".join(
            self.cli_search_path + ["/usr/local/bin", os.environ.get("PATH")]
        )
        cli_path = which(self.cli_name, mode=os.F_OK, path=env_path)
        if not cli_path:
            return None
        cli_path = which(cli_path, mode=os.F_OK, path=env_path)

        if cli_path:
            cli_path = Path(cli_path).resolve(strict=True)
            logger.debug(f"CLI found at {cli_path}")
        else:
            logger.debug(f"{self.cli_name} CLI not found.")

        return cli_path

    def get_version(self):
        """Invoke the manager and extract its own reported version.

        It does matter if this method return unsanitized and crappy string. The
        `version()` method below will clean and normalized it.
        """
        raise NotImplementedError

    @cachedproperty
    def version(self):
        """Parsed and normalized package manager's own version.

        Returns an instance of `TokenizedString`.
        """
        if self.executable:
            return self.get_version()

    @cachedproperty
    def supported(self):
        """ Is the package manager supported on that platform? """
        return CURRENT_OS_ID in self.platforms

    @cachedproperty
    def executable(self):
        """ Is the package manager CLI can be executed by the current user? """
        if not self.cli_path:
            return False
        if not os.access(self.cli_path, os.X_OK):
            logger.debug(f"{self.cli_path} not executable.")
            return False
        return True

    @cachedproperty
    def fresh(self):
        """ Does the package manager match the version requirement? """
        # Version is mandatory.
        if not self.version:
            return False
        if self.requirement:
            if self.version < parse_version(self.requirement):
                logger.debug(
                    f"{self.id} {self.version} is older than "
                    "{self.requirement} version requirement."
                )
                return False
        return True

    @cachedproperty
    def available(self):
        """Is the package manager available and ready-to-use on the system?

        Returns True only if the main CLI:
            1 - is supported on the current platform,
            2 - was found on the system,
            3 - is executable, and
            4 - match the version requirement.
        """
        return bool(self.supported and self.cli_path and self.executable and self.fresh)

    def run(self, *args, dry_run=False):
        """Run a shell command, return the output and keep error message.

        Removes ANSI escape codes, and returns ready-to-use strings.
        """
        # Serialize Path objects to strings.
        args = list(map(str, flatten(args)))
        args_str = click.style(" ".join(args), fg="white")
        logger.debug(f"► {args_str}")

        code = 0
        output = None
        error = None

        if not dry_run:
            code, output, error = run(*args)
        else:
            logger.warning("Dry-run: skip execution of command.")

        # Normalize messages.
        if error:
            error = strip_ansi(error)
            error = error if error else None
        if output:
            output = strip_ansi(output)
            output = output if output else None

        # Non-successful run.
        if code and error:
            # Produce an exception and eventually raise it.
            exception = CLIError(code, output, error)
            if self.raise_on_cli_error:
                raise exception
            # Accumulate errors.
            self.cli_errors.append(exception)

        # Log <stdout> and <stderr> output.
        if output:
            logger.debug(indent(output, "  "))
        if error:
            # Non-fatal error messages are logged as warnings.
            log_func = logger.error if code else logger.warning
            log_func(indent(error, "  "))

        return output

    def run_cli(self, *args, dry_run=False):
        """Like the `run` method above, but execute the binary pointed to by
        the `cli_path` property set in the current instance.
        """
        return self.run(self.cli_path, args, dry_run=dry_run)

    def sync(self):
        """ Refresh local manager metadata from remote repository. """
        if self.sync.__func__.__qualname__ == PackageManager.sync.__qualname__:
            logger.warning(f"Sync not implemented for {self.id}.")
            return
        logger.info(f"Sync {self.id} package info...")

    def cleanup(self):
        """ Remove left-overs and unused packages. """
        if self.cleanup.__func__.__qualname__ == PackageManager.cleanup.__qualname__:
            logger.warning(f"Cleanup not implemented for {self.id}.")
            return
        logger.info(f"Cleanup {self.id}...")

    @property
    def installed(self):
        """List packages currently installed on the system.

        Returns a dict indexed by package IDs. Each item is a dict with
        package ID, name and version.
        """
        raise NotImplementedError

    def search(self, query, extended, exact):
        """Search packages whose ID contain exact or partial query.

        Returns a dict indexed by package IDs. Each item is a dict with
        package ID, name, version and a boolean indicating if the match is
        exact or partial.
        """
        raise NotImplementedError

    @property
    def outdated(self):
        """List currently installed packages having a new version available.

        Returns a dict indexed by package IDs. Each item is a dict with
        package ID, name, current installed version and latest upgradeable
        version.
        """
        raise NotImplementedError

    def upgrade_cli(self, package_id=None):
        """ Return a bash-compatible full-CLI to upgrade a package. """
        raise NotImplementedError

    def upgrade(self, package_id=None, dry_run=False):
        """ Perform the upgrade of the provided package to latest version. """
        return self.run(self.upgrade_cli(package_id), dry_run=dry_run)

    def upgrade_all_cli(self):
        """ Return a bash-compatible full-CLI to upgrade all packages. """
        raise NotImplementedError

    def upgrade_all(self, dry_run=False):
        """Perform a full upgrade of all outdated packages to latest versions.

        If the manager doesn't implements a full upgrade one-liner, then
        fall-back to calling single-package upgrade one by one.
        """
        try:
            return self.run(self.upgrade_all_cli(), dry_run=dry_run)
        except NotImplementedError:
            logger.warning(
                f"Full upgrade subcommand not implemented in {self.id}. Call "
                "single-package upgrade CLI one by one."
            )
            log = []
            for package_id in self.outdated:
                output = self.upgrade(package_id, dry_run=dry_run)
                if output:
                    log.append(output)
            if log:
                return "\n".join(log)

    @staticmethod
    def render_cli(cmd, cli_format="plain"):
        """Return a formatted CLI in the requested format.

        * ``plain`` returns a simple string
        * ``fragments`` returns a list of strings
        * ``bitbar`` returns a CLI with parameters formatted into the bitbar
        dialect.
        """
        assert isinstance(cmd, list)
        assert cli_format in CLI_FORMATS
        cmd = map(str, flatten(cmd))  # Serialize Path instances.

        if cli_format == "fragments":
            return list(cmd)

        if cli_format == "plain":
            return " ".join(cmd)

        # Renders the CLI into BitBar dialect.
        bitbar_cli = ""
        for index, param in enumerate(cmd):
            if index == 0:
                bitbar_cli += "bash={}".format(param)
            else:
                if "=" in param:
                    param = '\\"{}\\"'.format(param)
                bitbar_cli += " param{}={}".format(index, param)
        return bitbar_cli



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import re

import simplejson as json

from ..base import PackageManager
from ..platform import LINUX, MACOS, WINDOWS
from ..version import parse_version


class Composer(PackageManager):

    name = "PHP's Composer"
    global_args = ["global"]
    platforms = frozenset([LINUX, MACOS, WINDOWS])
    requirement = "1.4.0"

    def get_version(self):
        """ Fetch version from ``composer --version`` output. """
        output = self.run_cli("--version")
        if output:
            return parse_version(output.split()[2])

    @property
    def installed(self):
        """Fetch installed packages from ``composer global show`` output.

        Raw CLI output samples:

        .. code-block:: shell-session

            ► composer global show --format=json | jq
            {
              "installed": [
                {
                  "name": "carbondate/carbon",
                  "version": "1.33.0",
                  "description": "A simple API extension for DateTime."
                },
                {
                  "name": "guzzlehttp/guzzle",
                  "version": "6.3.3",
                  "description": "Guzzle is a PHP HTTP client library"
                },
                {
                  "name": "guzzlehttp/promises",
                  "version": "v1.3.1",
                  "description": "Guzzle promises library"
                },
                {
                  "name": "guzzlehttp/psr7",
                  "version": "1.4.2",
                  "description": "PSR-7 message (...) methods"
                },
            (...)
        """
        installed = {}

        output = self.run_cli(self.global_args, "show", "--format=json")

        if output:

            package_list = json.loads(output)
            for package in package_list["installed"]:
                package_id = package["name"]
                installed[package_id] = {
                    "id": package_id,
                    "name": package_id,
                    "installed_version": parse_version(package["version"]),
                }

        return installed

    def search(self, query, extended, exact):
        """Fetch matching packages from ``composer search`` output.

        .. code-block:: shell-session

            ► composer global search symfony
            symfony/symfony The Symfony PHP framework
            symfony/yaml Symfony Yaml Component
            symfony/var-dumper Symfony (...) dumping PHP variables
            symfony/translation Symfony Translation Component
            symfony/routing Symfony Routing Component
            symfony/process Symfony Process Component
            symfony/polyfill-php70 Symfony (...) features to lower PHP versions
            symfony/polyfill-mbstring Symfony (...) Mbstring extension
            symfony/polyfill-ctype Symfony polyfill for ctype functions
            symfony/http-kernel Symfony HttpKernel Component
            symfony/http-foundation Symfony HttpFoundation Component
            symfony/finder Symfony Finder Component
            symfony/event-dispatcher Symfony EventDispatcher Component
            symfony/debug Symfony Debug Component
            symfony/css-selector Symfony CssSelector Component

        .. code-block:: shell-session

            ► composer global search --only-name python
            hiqdev/hidev-python
            aanro/pythondocx
            laravel-admin-ext/python-editor
            pythonphp/pythonphp
            blyxxyz/python-server
            nim-development/python-domotics
            rakshitbharat/pythoninphp
            tequilarapido/python-bridge

        .. code-block:: shell-session

            ► search global --only-name pythonphp/pythonphp
            pythonphp/pythonphp
        """
        matches = {}

        search_args = []
        if not extended:
            search_args.append("--only-name")

        output = self.run_cli(self.global_args, "search", search_args, query)

        if output:

            regexp = re.compile(
                r"""
                ^(?P<package_id>\S+\/\S+)
                (?P<description> .*)?
                """,
                re.MULTILINE | re.VERBOSE,
            )

            for package_id, description in regexp.findall(output):

                # Filters out fuzzy matches, only keep stricly matching
                # packages.
                if exact and query != package_id:
                    continue

                matches[package_id] = {
                    "id": package_id,
                    "name": package_id,
                    "latest_version": None,
                }

        return matches

    @property
    def outdated(self):
        """Fetch outdated packages from ``composer global outdated`` output.

        Raw CLI output samples:

        .. code-block:: shell-session

        ► composer global outdated --format=json
        {
            "installed": [
                {
                    "name": "illuminate/contracts",
                    "version": "v5.7.2",
                    "latest": "v5.7.3",
                    "latest-status": "semver-safe-update",
                    "description": "The Illuminate Contracts package."
                },
                {
                    "name": "illuminate/support",
                    "version": "v5.7.2",
                    "latest": "v5.7.3",
                    "latest-status": "semver-safe-update",
                    "description": "The Illuminate Support package."
                }
            ]
        }
        """
        outdated = {}

        output = self.run_cli(self.global_args, "outdated", "--format=json")

        if output:
            package_list = json.loads(output)
            for package in package_list["installed"]:
                package_id = package["name"]
                outdated[package_id] = {
                    "id": package_id,
                    "name": package_id,
                    "installed_version": parse_version(package["version"]),
                    "latest_version": parse_version(package["latest"]),
                }

        return outdated

    def upgrade_cli(self, package_id=None):
        cmd = [self.cli_path, self.global_args, "update"]
        if package_id:
            cmd.append(package_id)
        return cmd

    def upgrade_all_cli(self):
        return self.upgrade_cli()

    def cleanup(self):
        """Runs:

        .. code-block:: shell-session

            ► composer global clear-cache

        See: https://getcomposer.org/doc/03-cli.md#clear-cache-clearcache-cc
        """
        super(Composer, self).cleanup()
        self.run_cli(self.global_args, "clear-cache")



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import re

import simplejson as json
from boltons.cacheutils import cachedproperty

from ..base import PackageManager
from ..platform import MACOS
from ..version import parse_version
from . import logger


class Homebrew(PackageManager):

    """Virutal package manager shared by brew and cask CLI defined below.

    Homebrew is the umbrella project providing both brew and brew cask
    commands.
    """

    platforms = frozenset([MACOS])

    # Vanilla brew and cask CLIs now shares the same version.
    # 2.2.15 is the first release to support JSON output for outdated casks.
    requirement = "2.2.15"

    # Declare this manager as virtual, i.e. not tied to a real CLI.
    cli_name = None

    def get_version(self):
        """Fetch version from ``brew --version`` output.

        .. code-block:: shell-session

            ► brew --version
            Homebrew 1.8.6-124-g6cd4c31
            Homebrew/homebrew-core (git revision 533d; last commit 2018-12-28)
            Homebrew/homebrew-cask (git revision 5095b; last commit 2018-12-28)

        """
        output = self.run_cli("--version")
        if output:
            return parse_version(output.split()[1])

    def sync(self):
        """`brew` and `cask` share the same command.

        .. code-block:: shell-session

            ► brew update --quiet
            Already up-to-date.
        """
        super(Homebrew, self).sync()
        self.run_cli("update", "--quiet")

    @property
    def installed(self):
        """Fetch installed packages from ``brew list`` output.

        .. note::

            This method is shared by ``brew`` and ``cask``, only that the
            latter adds its ``cask`` subcommand to the CLI call.

        Raw CLI output samples:

        .. code-block:: shell-session

            ► brew list --versions
            ack 2.14
            apg 2.2.3
            audacity (!) 2.1.2
            apple-gcc42 4.2.1-5666.3
            atk 2.22.0
            bash 4.4.5
            bash-completion 1.3_1
            boost 1.63.0
            c-ares 1.12.0
            graphviz 2.40.1 2.40.20161221.0239
            quicklook-json latest

        .. code-block:: shell-session

            ► brew cask list --versions
            aerial 1.2beta5
            android-file-transfer latest
            audacity (!) 2.1.2
            bitbar 1.9.2
            firefox 49.0.1
            flux 37.7
            gimp 2.8.18-x86_64
            java 1.8.0_112-b16
            tunnelblick 3.6.8_build_4625 3.6.9_build_4685
            virtualbox 5.1.8-111374 5.1.10-112026

        Alternatives since 2.4.7 (see
        https://github.com/Homebrew/brew/pull/7949 and
        https://github.com/Homebrew/brew/pull/7966):

        .. code-block:: shell-session

            ► brew list --cask --versions
            aerial 1.2beta5
            android-file-transfer latest
            audacity (!) 2.1.2
            bitbar 1.9.2
            firefox 49.0.1
            flux 37.7
            gimp 2.8.18-x86_64
            java 1.8.0_112-b16
            tunnelblick 3.6.8_build_4625 3.6.9_build_4685
            virtualbox 5.1.8-111374 5.1.10-112026

            ► brew cask list --json --versions | jq
            [
              {
                "token": "aerial",
                "name": [
                  "Aerial Screensaver"
                ],
                "homepage": "https://github.com/JohnCoates/Aerial",
                "url": "https://github.com/(...)/download/v1.9.2/Aerial.zip",
                "appcast": "https://github.com/(...)/releases.atom",
                "version": "1.9.2",
                "sha256": "1d21511a31895ece4a18c93c779cbf4e35a611a27ba",
                "artifacts": [
                  [
                    "Aerial.saver"
                  ],
                  {
                    "trash": "~/Library/Caches/Aerial",
                    "signal": {}
                  }
                ],
                "caveats": null,
                "depends_on": {},
                "conflicts_with": null,
                "container": null,
                "auto_updates": null
              },
              {
                "token": "dropbox",
                "name": [
                  "Dropbox"
                ],
                "homepage": "https://www.dropbox.com/",
                "url": "https://www.dropbox.com/download?plat=mac&full=1",
                "appcast": null,
                "version": "latest",
                "sha256": "no_check",
                "artifacts": [
                  {
                    "launchctl": "com.dropbox.DropboxMacUpdate.agent",
                    "signal": {}
                  },
                  [
                    "Dropbox.app"
                  ],
                  {
                    "trash": [
                      "/Library/DropboxHelperTools",
                      "~/.dropbox",
                      "~/Library/Application Support/Dropbox",
                      "~/Library/Caches/com.dropbox.DropboxMacUpdate",
                      "~/Library/Caches/com.getdropbox.DropboxMetaInstaller",
                      "~/Library/Caches/com.getdropbox.dropbox",
                      "~/Library/Containers/com.dropbox.activityprovider",
                      "~/Library/Containers/com.dropbox.foldertagger",
                      "~/Library/Containers/com.getdropbox.dropbox.garcon",
                      "~/Library/Dropbox",
                      "~/Library/Group Containers/com.dropbox.client.crashpad",
                      "~/Library/Logs/Dropbox_debug.log",
                      "~/Library/Preferences/com.dropbox.DropboxMonitor.plist",
                      "~/Library/Preferences/com.getdropbox.dropbox.plist"
                    ],
                    "signal": {}
                  }
                ],
                "caveats": null,
                "depends_on": {},
                "conflicts_with": {
                  "cask": [
                    "dropbox-beta"
                  ]
                },
                "container": null,
                "auto_updates": null
              },
              (...)
            ]

        .. todo

            Use the ``removed`` variable to detect removed packages (which are
            reported with a ``(!)`` flag). See:
            https://github.com/caskroom/homebrew-cask/blob/master/doc
            /reporting_bugs/uninstall_wrongly_reports_cask_as_not_installed.md
            and https://github.com/kdeldycke/meta-package-manager/issues/17 .
        """
        installed = {}

        output = self.run_cli(self.global_args, "list", "--versions")

        if output:
            regexp = re.compile(r"(\S+)( \(!\))? (.+)")
            for pkg_info in output.splitlines():
                match = regexp.match(pkg_info)
                if match:
                    package_id, removed, versions = match.groups()
                    installed[package_id] = {
                        "id": package_id,
                        "name": package_id,
                        "installed_version":
                        # Keep highest version found.
                        max(map(parse_version, versions.split())),
                    }

        return installed

    def search(self, query, extended, exact):
        """Fetch matching packages from ``brew search`` output.

        .. code-block:: shell-session

            ► brew search sed
            ==> Formulae
            gnu-sed ✔                    libxdg-basedir               minised
            ==> Casks
            eclipse-dsl                       marsedit
            focused                           physicseditor
            google-adwords-editor             prefs-editor
            licensed                          subclassed-mnemosyne

        .. code-block:: shell-session

            ► brew search --formulae sed
            ==> Formulae
            gnu-sed ✔                    libxdg-basedir               minised

        .. code-block:: shell-session

            ► brew search --cask sed
            ==> Casks
            eclipse-dsl                       marsedit
            focused                           physicseditor
            google-adwords-editor             prefs-editor
            licensed                          subclassed-mnemosyne

        .. code-block:: shell-session

            ► brew search --formulae python
            ==> Formulae
            app-engine-python   boost-python3   python ✔          python-yq
            boost-python        gst-python      python-markdown   python@3.8 ✔

        .. code-block:: shell-session

            ► brew search --formulae "/^ssed$/"
            ==> Formulae
            ssed

        .. code-block:: shell-session

            ► brew search --formulae "/^sed$/"
            No formula or cask found for "/^sed$/".
            Closed pull requests:
            Merge ba7a794 (https://github.com/Homebrew/linuxbrew-core/pull/198)
            R: disable Tcl (https://github.com/Homebrew/homebrew-core/pull/521)
            (...)

        More doc at: https://docs.brew.sh/Manpage#search-texttext
        """
        matches = {}

        if extended:
            logger.warning(
                f"Extended search not supported for {self.id}. Fallback to Fuzzy."
            )

        # Use regexp for exact match.
        if exact:
            query = "/^{}$/".format(query)

        output = self.run_cli(self.search_args, query)

        if output:
            regexp = re.compile(
                r"""
                (?:==>\s\S+\s)?           # Ignore section starting with '==>'.
                (?P<package_id>[^\s✔]+)   # Anything not a whitespace or ✔.
                """,
                re.VERBOSE,
            )

            for package_id in regexp.findall(output):
                matches[package_id] = {
                    "id": package_id,
                    "name": package_id,
                    "latest_version": None,
                }

        return matches

    @property
    def outdated(self):
        """Fetch outdated packages from ``brew outdated`` output.

        Raw CLI output samples:

        .. code-block:: shell-session

            ► brew outdated --json=v1
            [
              {
                "name": "cassandra",
                "installed_versions": [
                  "3.5"
                ],
                "current_version": "3.7"
              },
              {
                "name": "vim",
                "installed_versions": [
                  "7.4.1967"
                ],
                "current_version": "7.4.1993"
              },
              {
                "name": "youtube-dl",
                "installed_versions": [
                  "2016.07.06"
                ],
                "current_version": "2016.07.09.1"
              }
            ]
        """
        outdated = {}

        # List available updates.
        output = self.run_cli(self.global_args, "outdated", "--json=v1")

        if output:
            for pkg_info in json.loads(output):
                package_id = pkg_info["name"]
                outdated[package_id] = {
                    "id": package_id,
                    "name": package_id,
                    "installed_version": max(
                        map(parse_version, pkg_info["installed_versions"])
                    ),
                    "latest_version": parse_version(pkg_info["current_version"]),
                }

        return outdated

    def upgrade_cli(self, package_id=None):
        """Returns the right CLI depending on weither formula or cask are
        concerned:

        .. code-block:: shell-session

            ► brew upgrade --formula
            ==> Upgrading 2 outdated packages:
            node 13.11.0 -> 13.12.0
            sdl2 2.0.12 -> 2.0.12_1
            ==> Upgrading node 13.11.0 -> 13.12.0
            ==> Downloading https://homebrew.bintray.com/bottles/node-13.tar.gz
            ==> Downloading from https://akamai.bintray.com/fc/fc0bfb42fe23e960
            ############################################################ 100.0%
            ==> Pouring node-13.12.0.catalina.bottle.tar.gz
            ==> Caveats
            Bash completion has been installed to:
              /usr/local/etc/bash_completion.d
            ==> Summary
            🍺  /usr/local/Cellar/node/13.12.0: 4,660 files, 60.3MB
            Removing: /usr/local/Cellar/node/13.11.0... (4,686 files, 60.4MB)
            ==> Upgrading sdl2 2.0.12 -> 2.0.12_1
            ==> Downloading https://homebrew.bintray.com/bottles/sdl2-2.tar.gz
            ==> Downloading from https://akamai.bintray.com/4d/4dcd635465d16372
            ############################################################ 100.0%
            ==> Pouring sdl2-2.0.12_1.catalina.bottle.tar.gz
            🍺  /usr/local/Cellar/sdl2/2.0.12_1: 89 files, 4.7MB
            Removing: /usr/local/Cellar/sdl2/2.0.12... (89 files, 4.7MB)
            ==> Checking for dependents of upgraded formulae...
            ==> No dependents found!
            ==> Caveats
            ==> node
            Bash completion has been installed to:
              /usr/local/etc/bash_completion.d

        .. code-block:: shell-session

            ► brew upgrade --cask
            Updating Homebrew...
            ==> Auto-updated Homebrew!
            Updated Homebrew from 1654de327 to cfa03c8cc.
            Updated 2 taps (homebrew/core and homebrew/cask).
            ==> Casks with `auto_updates` or `version :latest` will not be upgraded
            ==> Upgrading 1 outdated packages:
            aerial 2.0.7 -> 2.0.8
            ==> Upgrading aerial
            ==> Downloading https://github.com/Aerial/download/v2.0.8/Aerial.saver.zip
            ==> Downloading from https://65be.s3.amazonaws.com/44998092/29eb1e0
            ==> Verifying SHA-256 checksum for Cask 'aerial'.
            ==> Backing Screen Saver up to '/usr/local/Caskroom/Aerial.saver'.
            ==> Removing Screen Saver '/Users/kde/Library/Screen Savers/Aerial.saver'.
            ==> Moving Screen Saver to '/Users/kde/Library/Screen Savers/Aerial.saver'.
            ==> Purging files for version 2.0.7 of Cask aerial
            🍺  aerial was successfully upgraded!
        """
        cmd = [self.cli_path] + self.upgrade_args
        if package_id:
            cmd.append(package_id)
        return cmd

    def upgrade_all_cli(self):
        return self.upgrade_cli()

    def cleanup(self):
        """Scrub the cache, including downloads for even the latest versions.

        Note downloads for any installed formulae or casks will still not be
        deleted.

        .. code-block:: shell-session

            ► brew cleanup -s
            Removing: ~/Library/Caches/Homebrew/node--1.bottle.tar.gz... (9MB)
            Warning: Skipping sdl2: most recent version 2.0.12_1 not installed
            Removing: ~/Library/Caches/Homebrew/Cask/aerial--1.8.1.zip... (5MB)
            Removing: ~/Library/Caches/Homebrew/Cask/prey--1.9.pkg... (19.9MB)
            Removing: ~/Library/Logs/Homebrew/readline... (64B)
            Removing: ~/Library/Logs/Homebrew/libfido2... (64B)
            Removing: ~/Library/Logs/Homebrew/libcbor... (64B)

        More doc at: https://docs.brew.sh/Manpage#cleanup-options-formulacask
        """
        super(Homebrew, self).cleanup()
        self.run_cli("cleanup", "-s")


class Brew(Homebrew):

    name = "Homebrew Formulae"
    cli_name = "brew"

    @cachedproperty
    def search_args(self):
        """Returns arguments needed for search of Homebrew formulae.

        .. code-block:: shell-session

            ► brew search --formulae sed
            ==> Formulae
            gnu-sed ✔                    libxdg-basedir               minised
        """
        return [self.global_args, "search", "--formulae"]

    @cachedproperty
    def upgrade_args(self):
        """Returns arguments needed for upgrade of Homebrew formulae.

        .. code-block:: shell-session

            ► brew upgrade --formula
            ==> Upgrading 2 outdated packages:
            node 13.11.0 -> 13.12.0
            sdl2 2.0.12 -> 2.0.12_1
            ==> Upgrading node 13.11.0 -> 13.12.0
            ==> Downloading https://homebrew.bintray.com/bottles/node-13.tar.gz
            ==> Downloading from https://akamai.bintray.com/fc/fc0bfb42fe23e960
            ############################################################ 100.0%
            ==> Pouring node-13.12.0.catalina.bottle.tar.gz
            ==> Caveats
            Bash completion has been installed to:
              /usr/local/etc/bash_completion.d
            ==> Summary
            🍺  /usr/local/Cellar/node/13.12.0: 4,660 files, 60.3MB
            Removing: /usr/local/Cellar/node/13.11.0... (4,686 files, 60.4MB)
            ==> Upgrading sdl2 2.0.12 -> 2.0.12_1
            ==> Downloading https://homebrew.bintray.com/bottles/sdl2-2.tar.gz
            ==> Downloading from https://akamai.bintray.com/4d/4dcd635465d16372
            ############################################################ 100.0%
            ==> Pouring sdl2-2.0.12_1.catalina.bottle.tar.gz
            🍺  /usr/local/Cellar/sdl2/2.0.12_1: 89 files, 4.7MB
            Removing: /usr/local/Cellar/sdl2/2.0.12... (89 files, 4.7MB)
            ==> Checking for dependents of upgraded formulae...
            ==> No dependents found!
            ==> Caveats
            ==> node
            Bash completion has been installed to:
              /usr/local/etc/bash_completion.d
        """
        return [self.global_args, "upgrade", "--formula"]


class Cask(Homebrew):

    """Cask is now part of Homebrew's core and extend it."""

    name = "Homebrew Cask"
    cli_name = "brew"

    global_args = ["cask"]

    @cachedproperty
    def search_args(self):
        """Returns arguments needed for search of Homebrew casks.

        .. code-block:: shell-session

            ► brew search --cask sed
            ==> Casks
            eclipse-dsl                       marsedit
            focused                           physicseditor
            google-adwords-editor             prefs-editor
            licensed                          subclassed-mnemosyne
        """
        return ["search", "--cask"]

    @property
    def outdated(self):
        """Search for outdated packages among installed one.

        .. code-block:: shell-session

            ► brew cask outdated --json | jq
            [
              {
                "name": "google-play-music-desktop-player",
                "installed_versions": "4.4.0",
                "current_version": "4.4.1"
              },
              {
                "name": "prey",
                "installed_versions": "1.8.3.upgrading",
                "current_version": "1.9.3"
              }
            ]

        .. code-block:: shell-session

            ► brew cask outdated --json --greedy | jq u
            [
              {
                "name": "android-file-transfer",
                "installed_versions": "latest",
                "current_version": "latest"
              },
              {
                "name": "atom",
                "installed_versions": "1.19.3",
                "current_version": "1.19.4"
              },
              {
                "name": "keybase",
                "installed_versions": "4.0.0-20190507193726,6614a49937",
                "current_version": "5.5.0-20200526170801,139bb348af"
              }
            ]
        """
        outdated = {}

        # Build up the list of CLI options.
        options = ["--json"]
        # Includes auto-update packages or not.
        if not self.ignore_auto_updates:
            options.append("--greedy")

        # List available updates.
        output = self.run_cli(self.global_args, "outdated", options)

        if output:
            for pkg_info in json.loads(output):
                package_id = pkg_info["name"]
                version = pkg_info["installed_versions"]
                latest_version = pkg_info["current_version"]

                # Skip packages in undetermined state.
                if version == latest_version:
                    continue

                outdated[package_id] = {
                    "id": package_id,
                    "name": package_id,
                    "installed_version": parse_version(version),
                    "latest_version": parse_version(latest_version),
                }

        return outdated

    @cachedproperty
    def upgrade_args(self):
        """Returns arguments needed for upgrade of Homebrew casks.

        .. code-block:: shell-session

            ► brew upgrade --cask
            Updating Homebrew...
            ==> Auto-updated Homebrew!
            Updated Homebrew from 1654de327 to cfa03c8cc.
            Updated 2 taps (homebrew/core and homebrew/cask).
            ==> Casks with `auto_updates` or `version :latest` will not be upgraded
            ==> Upgrading 1 outdated packages:
            aerial 2.0.7 -> 2.0.8
            ==> Upgrading aerial
            ==> Downloading https://github.com/Aerial/download/v2.0.8/Aerial.saver.zip
            ==> Downloading from https://65be.s3.amazonaws.com/44998092/29eb1e0
            ==> Verifying SHA-256 checksum for Cask 'aerial'.
            ==> Backing Screen Saver up to '/usr/local/Caskroom/Aerial.saver'.
            ==> Removing Screen Saver '/Users/kde/Library/Screen Savers/Aerial.saver'.
            ==> Moving Screen Saver to '/Users/kde/Library/Screen Savers/Aerial.saver'.
            ==> Purging files for version 2.0.7 of Cask aerial
            🍺  aerial was successfully upgraded!
        """
        return ["upgrade", "--cask"]



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import simplejson as json
from boltons.cacheutils import cachedproperty

from ..base import PackageManager
from ..platform import LINUX, MACOS, WINDOWS
from ..version import parse_version
from . import logger


class Yarn(PackageManager):

    name = "Node's yarn"

    global_args = []

    platforms = frozenset([LINUX, MACOS, WINDOWS])

    requirement = "1.21.0"

    def get_version(self):
        """Fetch version from ``yarn --version`` output.

        .. code-block:: shell-session

            ► yarn --version
            (...)
        """
        return parse_version(self.run_cli("--version"))

    def parse(self, output):
        packages = {}

        if not output:
            return packages

        for line in output.splitlines():
            if not line:
                continue
            obj = json.loads(line)
            if obj["type"] != "info":
                continue
            package = self.parse_info(obj)
            packages[package["id"]] = package
        return packages

    @staticmethod
    def parse_info(obj):
        data = obj["data"].replace("has binaries:", "")
        parts = data.replace('"', "").split("@")
        package_id = parts[0]
        version = parts[1]
        return {
            "id": package_id,
            "name": package_id,
            "installed_version": parse_version(version),
        }

    @property
    def installed(self):
        """Fetch installed packages from ``yarn list`` output.

        .. code-block:: shell-session

            ► yarn global --json list --depth 0
            (...)
        """
        output = self.run_cli(
            "global", "--json", self.global_args, "list", "--depth", "0"
        )
        return self.parse(output)

    def search(self, query, extended, exact):
        """Call ``yarn info`` and only works for exact match.

        Yarn maintainers have decided not to implement a dedicated ``search``
        command:
        https://github.com/yarnpkg/yarn/issues/778#issuecomment-253146299

        .. code-block:: shell-session

            ► yarn --json info python | jq
            {
              "type": "inspect",
              "data": {
                "name": "python",
                "description": "Interact with python child process",
                "dist-tags": {
                  "latest": "0.0.4"
                },
                "versions": [
                  "0.0.0",
                  "0.0.1",
                  "0.0.2",
                  "0.0.3",
                  "0.0.4"
                ],
                "maintainers": [
                  {
                    "name": "drderidder",
                    "email": "drderidder@gmail.com"
                  }
                ],
                "time": {
                  "modified": "2017-09-16T05:26:13.151Z",
                  "created": "2011-07-11T01:59:04.362Z",
                  "0.0.0": "2011-07-11T01:59:05.137Z",
                  "0.0.1": "2011-07-17T05:23:33.166Z",
                  "0.0.2": "2011-07-20T03:42:50.379Z",
                  "0.0.3": "2014-06-08T00:39:08.562Z",
                  "0.0.4": "2015-01-25T02:48:07.820Z"
                },
                "author": {
                  "name": "Darren DeRidder"
                },
                "repository": {
                  "type": "git",
                  "url": "git://github.com/73rhodes/node-python.git"
                },
                "homepage": "https://github.com/73rhodes/node-python",
                "bugs": {
                  "url": "https://github.com/73rhodes/node-python/issues"
                },
                "readmeFilename": "README.md",
                "users": {
                  "dewang-mistry": true,
                  "goliatone": true,
                  "sapanbhuta": true,
                  "aditcmarix": true,
                  "imlucas": true,
                  "heyderpd": true,
                  "ukuli": true,
                  "chbardel": true,
                  "asaupup": true,
                  "nuwaio": true
                },
                "version": "0.0.4",
                "main": "./lib/python.js",
                "engines": {
                  "node": ">= 0.4.1"
                },
                "gitHead": "69754aaa57658193916a1bf5fc391198098f74f6",
                "scripts": {},
                "dist": {
                  "shasum": "3094e898ef17a33aa9c3e973b3848a38e47d1818",
                  "tarball": "https://registry.npmjs.org/python/-/python-1.tgz"
                },
                "directories": {}
              }
            }
        """
        matches = {}

        if extended:
            logger.warning(
                f"Extended search not supported for {self.id}. Fallback to Fuzzy."
            )
        elif not exact:
            logger.warning(
                f"Fuzzy search not supported for {self.id}. Fallback to Exact."
            )

        output = self.run_cli(self.global_args, "--json", "info", query)

        if output:
            result = json.loads(output)

            if result["type"] == "inspect":
                package = result["data"]
                package_id = package["name"]
                matches[package_id] = {
                    "id": package_id,
                    "name": package_id,
                    "latest_version": parse_version(package["version"]),
                }

        return matches

    @cachedproperty
    def global_dir(self):
        return self.run_cli("global", "dir").rstrip()

    @property
    def outdated(self):
        """Fetch outdated packages from ``yarn outdated`` output.

        Raw CLI output samples:

        .. code-block:: shell-session

            ► yarn --json outdated --cwd
            (...)
        """
        outdated = {}

        output = self.run_cli(
            "--json", self.global_args, "outdated", "--cwd", self.global_dir
        )

        if not output:
            return outdated

        packages = list()
        for line in output.splitlines():
            if not line:
                continue
            obj = json.loads(line)
            if obj["type"] == "table":
                packages = obj["data"]["body"]
                break

        for package in packages:
            package_id = package[0]
            values = {"current": package[1], "wanted": package[2], "latest": package[3]}

            if values["wanted"] == "linked":
                continue
            outdated[package_id] = {
                "id": package_id + "@" + values["latest"],
                "name": package_id,
                "installed_version": parse_version(values["current"]),
                "latest_version": parse_version(values["latest"]),
            }
        return outdated

    def upgrade_cli(self, package_id=None):
        cmd = [self.cli_path, "global", self.global_args]

        if package_id:
            cmd.append("add")
            cmd.append(package_id)
        else:
            cmd.append("upgrade")

        return cmd

    def upgrade_all_cli(self):
        return self.upgrade_cli()

    def cleanup(self):
        """Remove the shared cache files.

        See: https://yarnpkg.com/cli/cache/clean
        """
        super(Yarn, self).cleanup()
        self.run_cli(self.global_args, "cache", "clean", "--all")



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import re
from pathlib import Path, PurePath
from types import MethodType

from ..managers import pool
from ..platform import OS_DEFINITIONS
from ..version import TokenizedString
from .conftest import MANAGER_IDS

""" Test the structure, data and types returned by all package managers. """


def test_manager_count():
    """ Check all implemented package managers are accounted for. """
    assert len(pool()) == 13
    assert len(pool()) == len(MANAGER_IDS)
    assert MANAGER_IDS == set(pool())


def test_cached_pool():
    assert pool() == pool()
    assert pool() is pool()


def test_sorted_pool():
    assert list(pool()) == sorted([m.id for m in pool().values()])


def test_ascii_id():
    """ All package manager IDs should be short ASCII strings. """
    for manager_id, manager in pool().items():
        assert manager_id
        assert isinstance(manager_id, str)
        assert manager_id.isascii()
        assert manager_id.isalnum()
        assert manager_id.islower()
        assert manager_id == manager.id


def test_name():
    """ Check all managers have a name. """
    name_matcher = re.compile(r"[a-zA-Z0-9\' ]+")
    for manager in pool().values():
        assert manager.name
        assert isinstance(manager.name, str)
        assert name_matcher.match(manager.name)
    # Names are unique.
    assert len({m.name for m in pool().values()}) == len(MANAGER_IDS)


def test_platforms():
    """Check that definitions returns supported platforms as a frozenset."""
    for manager in pool().values():
        assert manager.platforms
        assert isinstance(manager.platforms, frozenset)
        assert manager.platforms.issubset(OS_DEFINITIONS)


def test_cli_name_type():
    """ Check the pointed CLI name and path are file-system compatible. """
    for manager in pool().values():
        assert manager.cli_name
        assert isinstance(manager.cli_name, str)
        assert manager.cli_name.isalnum()
        assert PurePath(manager.cli_name).name == manager.cli_name


def test_virtual():
    """ Check the manager as a defined virtual property. """
    for manager in pool().values():
        assert isinstance(manager.virtual, bool)


def test_cli_search_path():
    for manager in pool().values():
        assert isinstance(manager.cli_search_path, list)
        assert len(set(manager.cli_search_path)) == len(manager.cli_search_path)
        for search_path in manager.cli_search_path:
            assert isinstance(search_path, str)
            path_obj = Path(search_path).resolve()
            assert path_obj.is_absolute()
            assert not path_obj.is_reserved()
            if path_obj.exists():
                assert path_obj.is_file() or path_obj.is_dir()


def test_cli_path():
    for manager in pool().values():
        if manager.cli_path is not None:
            assert isinstance(manager.cli_path, Path)
            assert manager.cli_path.is_absolute()
            assert not manager.cli_path.is_reserved()
            assert manager.cli_path.is_file()


def test_global_args_type():
    """ Check that definitions returns CLI args as a list of strings. """
    arg_matcher = re.compile(r"[a-zA-Z0-9-]+")
    for manager in pool().values():
        assert isinstance(manager.global_args, list)
        for arg in manager.global_args:
            assert arg
            assert isinstance(arg, str)
            assert arg_matcher.match(arg)


def test_requirement():
    """ Each manager is required to specify a minimal version. """
    req_matcher = re.compile(r"[0-9\.]+")
    for manager in pool().values():
        assert isinstance(manager.requirement, str)
        assert req_matcher.match(manager.requirement)
        # Check provided string is lossless once passed via TokenizedString.
        assert str(TokenizedString(manager.requirement)) == manager.requirement


def test_get_version():
    """ Check that method parsing the CLI version returns a string. """
    for manager in pool().values():
        assert isinstance(manager.get_version, MethodType)
        if manager.executable:
            if manager.get_version() is not None:
                assert isinstance(manager.get_version(), TokenizedString)


def test_version():
    for manager in pool().values():
        if manager.version is not None:
            assert isinstance(manager.version, TokenizedString)


def test_supported():
    for manager in pool().values():
        assert isinstance(manager.supported, bool)


def test_executable():
    for manager in pool().values():
        assert isinstance(manager.executable, bool)


def test_fresh():
    for manager in pool().values():
        assert isinstance(manager.fresh, bool)


def test_available():
    for manager in pool().values():
        assert isinstance(manager.available, bool)


def test_cli_type():
    """ Check that all methods returning a CLI is a list. """
    for manager in pool().values():
        assert isinstance(manager.upgrade_cli("dummy_package_id"), list)

        # Upgrade-all CLI are allowed to raise a particular error.
        try:
            result = manager.upgrade_all_cli()
        except Exception as excpt:
            assert isinstance(excpt, NotImplementedError)
        else:
            assert isinstance(result, list)


def test_installed_type():
    """ Check that all installed operations returns a dict of dicts. """
    for manager in pool().values():
        if manager.available:
            assert isinstance(manager.installed, dict)
            for pkg in manager.installed.values():
                assert isinstance(pkg, dict)
                assert set(pkg) == {"id", "name", "installed_version"}
                assert isinstance(pkg["id"], str)
                assert isinstance(pkg["name"], str)
                if pkg["installed_version"] is not None:
                    assert isinstance(pkg["installed_version"], TokenizedString)


def test_search_type():
    """ Check that all search operations returns a dict of dicts. """
    for manager in pool().values():
        if manager.available:
            matches = manager.search("python", extended=True, exact=False)
            assert isinstance(matches, dict)
            for pkg in matches.values():
                assert isinstance(pkg, dict)
                assert set(pkg) == {"id", "name", "latest_version"}
                assert isinstance(pkg["id"], str)
                assert isinstance(pkg["name"], str)
                if pkg["latest_version"] is not None:
                    assert isinstance(pkg["latest_version"], TokenizedString)


def test_outdated_type():
    """ Check that all outdated operations returns a dict of dicts. """
    for manager in pool().values():
        if manager.available:
            assert isinstance(manager.outdated, dict)
            for pkg in manager.outdated.values():
                assert isinstance(pkg, dict)
                assert set(pkg) == {"id", "name", "installed_version", "latest_version"}
                assert isinstance(pkg["id"], str)
                assert isinstance(pkg["name"], str)
                if pkg["installed_version"] is not None:
                    assert isinstance(pkg["installed_version"], TokenizedString)
                assert isinstance(pkg["latest_version"], TokenizedString)


def test_sync_type():
    """ Check that sync operation return nothing. """
    for manager in pool().values():
        if manager.available:
            assert isinstance(manager.sync, MethodType)
            assert manager.sync() is None


def test_cleanup_type():
    """ Check that cleanup operation return nothing. """
    for manager in pool().values():
        if manager.available:
            assert isinstance(manager.cleanup, MethodType)
            assert manager.cleanup() is None



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

""" Utilities to load parameters and options from a configuration file. """

import click
import tomlkit
from pathlib import Path
from . import logger, CLI_NAME


def config_structure():
    """Returns the supported configuration structure.

    Derives TOML structure from CLI definition.
    """
    # Imported here to avoid circular imports.
    from .cli import cli

    # Global, top-level options shared by all subcommands are placed under the
    # cli name's section.
    # TODO: exclude version option.
    # TODO: take particular attention to boolean flag options.
    config = {CLI_NAME: {p.name for p in cli.params}}

    # Subcommand-specific options.
    for cmd_id, cmd in cli.commands.items():
        config[cmd_id] = {p.name for p in cmd.params}

    return config


def read_config(custom_conf=None):
    """ Loads a configuration files and returns recognized options.

    If no config file provided, defaults to the ``config.toml`` file found in a
    folder depending on the OS:
    * macOS & Linux:
        ~/.mpm/
    * Windows XP:
        C:\\Documents and Settings\\<user>\\Local Settings\\Application Data\\mpm\\
    * Windows 7:
        C:\\Users\\<user>\\AppData\\Roaming\\mpm\\

    As per: https://click.palletsprojects.com/en/7.x/api/#click.get_app_dir

    Invalid parameters are ignored and log messages are emitted.
    """
    # The recognized configuration extracted from the file.
    valid_config = {}

    # Get configuration from provided file.
    if custom_conf:
        cfg_filepath = Path(custom_conf).resolve()
    # Get configuration from default location.
    else:
        cfg_filepath = Path(
            click.get_app_dir(CLI_NAME, force_posix=True), "config.toml"
        ).resolve()

    # Check config file. Issues with non-default config file are fatal.
    if not cfg_filepath.exists():
        if custom_conf:
            logger.fatal(f"Configuration not found at {cfg_filepath}")
            raise FileNotFoundError
        else:
            logger.warning(f"Configuration not found at {cfg_filepath}")
            return valid_config
    if not cfg_filepath.is_file():
        logger.fatal(f"Configuration {cfg_filepath} is not a file.")
        raise FileNotFoundError

    # Parse TOML content.
    logger.info(f"Load configuration from {cfg_filepath}")
    doc = tomlkit.parse(cfg_filepath.read_bytes())

    # Filters-out configuration file's content against the reference structure and
    # only keep recognized options.
    structure = config_structure()

    for section, options in doc.items():

        # Ignore unrecognized section.
        if section not in structure:
            logger.warning(f"Ignore [{section}] section.")
            continue

        # Dive into section and look at options.
        valid_options = {}
        for opt_id, opt_value in options.items():

            # Ignore unrecognized option.
            if opt_id not in structure[section]:
                logger.warning(f"Ignore [{section}].{opt_id} option.")
                continue

            # Keep option.
            valid_options[opt_id] = opt_value

        valid_config[section] = valid_options

    return valid_config



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

# pylint: disable=redefined-outer-name

import re

import pytest
import simplejson as json
from boltons.iterutils import same

from .conftest import MANAGER_IDS, unless_macos
from .test_cli import CLISubCommandTests, CLITableTests


@pytest.fixture
def subcmd():
    return "search", "abc"


class TestSearch(CLISubCommandTests, CLITableTests):
    @pytest.mark.parametrize("mid", MANAGER_IDS)
    def test_single_manager(self, invoke, subcmd, mid):
        result = invoke("--manager", mid, subcmd)
        assert result.exit_code == 0
        self.check_manager_selection(result, {mid})

    def test_json_parsing(self, invoke, subcmd):
        result = invoke("--output-format", "json", subcmd)
        assert result.exit_code == 0
        data = json.loads(result.stdout)

        assert data
        assert isinstance(data, dict)
        assert set(data).issubset(MANAGER_IDS)

        for manager_id, info in data.items():
            assert isinstance(manager_id, str)
            assert isinstance(info, dict)

            assert isinstance(info["id"], str)
            assert isinstance(info["name"], str)

            assert set(info) == {"errors", "id", "name", "packages"}

            assert isinstance(info["errors"], list)
            if info["errors"]:
                assert same(map(type, info["errors"]), str)

            assert info["id"] == manager_id

            assert isinstance(info["packages"], list)
            for pkg in info["packages"]:
                assert isinstance(pkg, dict)

                assert set(pkg) == {"id", "latest_version", "name"}

                assert isinstance(pkg["id"], str)
                if pkg["latest_version"] is not None:
                    assert isinstance(pkg["latest_version"], str)
                assert isinstance(pkg["name"], str)

    @unless_macos
    def test_unicode_search(self, invoke):
        """ See #16. """
        result = invoke("--manager", "cask", "search", "ubersicht")
        assert result.exit_code == 0
        assert "ubersicht" in result.stdout
        # XXX search command is not fetching yet detailed package infos like
        # names.
        assert "Übersicht" not in result.stdout

        result = invoke("--manager", "cask", "search", "Übersicht")
        assert result.exit_code == 0
        assert "ubersicht" in result.stdout
        assert "Übersicht" not in result.stdout

    def test_exact_search_tokenizer_one_result(self, invoke):
        result = invoke("--manager", "pip", "search", "--exact", "sed", color=False)
        assert result.exit_code == 0
        assert "1 package total" in result.stdout
        assert " sed " in result.stdout

    @pytest.mark.parametrize(
        "query", ["SED", "SeD", "sEd*", "*sED*", "_seD-@", "", "_"]
    )
    def test_exact_search_tokenizer_no_result(self, invoke, query):
        result = invoke("--manager", "pip", "search", "--exact", query)
        assert result.exit_code == 0
        assert "0 package total" in result.stdout
        assert "sed" not in result.stdout

    @pytest.mark.parametrize("query", ["", "_", "_seD-@"])
    def test_fuzzy_search_tokenizer_no_results(self, invoke, query):
        result = invoke("--manager", "pip", "search", query)
        assert result.exit_code == 0
        assert "0 package total" in result.stdout
        assert "sed" not in result.stdout

    @pytest.mark.parametrize("query", ["sed", "SED", "SeD", "sEd*", "*sED*"])
    def test_fuzzy_search_tokenizer_multiple_results(self, invoke, query):
        result = invoke("--manager", "pip", "search", query, color=False)
        assert result.exit_code == 0
        assert "2 packages total" in result.stdout
        assert " sed " in result.stdout
        assert " SED-cli " in result.stdout

    @pytest.mark.parametrize("query", ["", "_", "_seD-@"])
    def test_extended_search_tokenizeri_no_results(self, invoke, query):
        result = invoke("--manager", "pip", "search", "--extended", query)
        assert result.exit_code == 0
        assert "0 package total" in result.stdout
        assert "sed" not in result.stdout

    @pytest.mark.parametrize("query", ["sed", "SED", "SeD", "sEd*", "*sED*"])
    def test_extended_search_tokenizeri_multiple_results(self, invoke, query):
        result = invoke("--manager", "pip", "search", "--extended", query)
        assert result.exit_code == 0
        last_line = result.stdout.splitlines()[-1]
        assert last_line
        msg_match = re.match(
            r"^([0-9]+) packages? total \(pip: ([0-9]+)\).$", last_line
        )
        assert msg_match
        assert same(msg_match.groups())
        # We should find lots of results for this package search.
        assert int(msg_match.groups()[0]) >= 20



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import re

from ..base import PackageManager
from ..platform import LINUX, MACOS, WINDOWS
from ..version import parse_version
from . import logger


class Gem(PackageManager):

    platforms = frozenset([LINUX, MACOS, WINDOWS])

    name = "Ruby Gems"

    # Default to the version shipped with the latest maintained macOS version,
    # i.e. macOS 10.13 High Sierra, which is bundled with gem 2.5.2.
    requirement = "2.5.0"

    # Help mpm a little bit in its search for the `gem` binary.
    cli_search_path = ["/usr/local/opt/ruby/bin/gem", "/usr/local/opt/ruby/bin"]

    global_args = [
        "--quiet",  # Silence command progress meter
    ]

    def get_version(self):
        """Fetch version from ``gem --version`` output.

        Raw CLI output samples:

        .. code-block:: shell-session

            ► gem --version
            3.0.3
        """
        return parse_version(self.run_cli("--version"))

    @property
    def installed(self):
        """Fetch installed packages from ``gem list`` output.

        Raw CLI output samples:

        .. code-block:: shell-session

            ► gem list --quiet
            bigdecimal (default: 1.4.1)
            bundler (default: 1.17.2)
            CFPropertyList (2.3.6)
            cmath (default: 1.0.0)
            csv (default: 3.0.9)
            date (default: 2.0.0)
            fileutils (1.4.1, default: 1.1.0)
            io-console (0.5.6, default: 0.4.7)
            ipaddr (default: 1.2.2)
            molinillo (0.5.4, 0.4.5, 0.2.3)
            nokogiri (1.5.6)
            psych (2.0.0)
            rake (0.9.6)
            rdoc (4.0.0)
            sqlite3 (1.3.7)
            test-unit (2.0.0.0)
        """
        installed = {}

        output = self.run_cli("list", self.global_args)

        if output:
            regexp = re.compile(r"(\S+) \((.+)\)")
            for package in output.splitlines():
                match = regexp.match(package)
                if match:
                    package_id, versions = match.groups()
                    # Guess latest installed version.
                    version = max(
                        [
                            parse_version(v)
                            for v in re.compile(r",|default:| ").split(versions)
                            if v
                        ]
                    )
                    installed[package_id] = {
                        "id": package_id,
                        "name": package_id,
                        "installed_version": version,
                    }

        return installed

    def search(self, query, extended, exact):
        """Fetch matching packages from ``gem search`` output.

        .. code-block:: shell-session

            ► gem search python --versions --quiet
            at_coder_friends-generator-python_ref (0.2.0)
            bee_python (0.2.3)
            dependabot-python (0.117.5)
            logstash-filter-python (0.0.1 java)
            python (0.0.1)
            python-generator (1.1.0)
            python_with_git_test (2.499.8)
            rabbit-slide-niku-erlangvm-for-pythonista (2015.09.12)
            RubyToPython (0.0)

        .. code-block:: shell-session

            ► gem search python --versions --exact --quiet
            python (0.0.1)
        """
        matches = {}

        if extended:
            logger.warning(
                f"Extended search not supported for {self.id}. Fallback to Fuzzy."
            )

        search_arg = []
        if exact:
            search_arg.append("--exact")

        output = self.run_cli(
            "search", query, "--versions", search_arg, self.global_args
        )

        if output:
            regexp = re.compile(
                r"""
                (?P<package_id>\S+)     # Any string.
                \                       # A space.
                \(                      # Start of content in parenthesis.
                    (?P<version>\S+)    # Version string.
                    (?:\ \S+)?          # Optional platform value after space.
                \)
                """,
                re.VERBOSE,
            )
            for package_id, version in regexp.findall(output):
                matches[package_id] = {
                    "id": package_id,
                    "name": package_id,
                    "latest_version": parse_version(version),
                }

        return matches

    @property
    def outdated(self):
        """Fetch outdated packages from ``gem outdated`` output.

        Raw CLI output samples:

        .. code-block:: shell-session

            ► gem outdated
            did_you_mean (1.0.0 < 1.0.2)
            io-console (0.4.5 < 0.4.6)
            json (1.8.3 < 2.0.1)
            minitest (5.8.3 < 5.9.0)
            power_assert (0.2.6 < 0.3.0)
            psych (2.0.17 < 2.1.0)
        """
        outdated = {}

        output = self.run_cli("outdated", self.global_args)

        if output:
            regexp = re.compile(r"(\S+) \((\S+) < (\S+)\)")
            for package in output.splitlines():
                match = regexp.match(package)
                if match:
                    package_id, current_version, latest_version = match.groups()
                    outdated[package_id] = {
                        "id": package_id,
                        "name": package_id,
                        "installed_version": parse_version(current_version),
                        "latest_version": parse_version(latest_version),
                    }

        return outdated

    def upgrade_cli(self, package_id=None):
        cmd = [self.cli_path, "update", "--user-install", self.global_args]
        # Installs require `sudo` on system ruby.
        # I (@tresni) recommend doing something like:
        #     ► sudo dseditgroup -o edit -a -t user wheel
        # And then do `visudo` to make it so the `wheel` group does not require
        # a password. There is a line already there for it, you just need to
        # uncomment it and save.)
        # if self.cli_path == '/usr/bin/gem':
        #     cmd.insert(0, '/usr/bin/sudo')
        if package_id:
            cmd.append(package_id)
        return cmd

    def upgrade_all_cli(self):
        return self.upgrade_cli()

    def cleanup(self):
        """ Run ``gem cleanup`` CLI.

        Raw CLI output samples:

        .. code-block:: shell-session

            ► gem cleanup
            Cleaning up installed gems...
            Attempting to uninstall test-unit-3.2.9
            Unable to uninstall test-unit-3.2.9:
                Gem::FilePermissionError: You don't have write permissions \
                for the /Library/Ruby/Gems/2.6.0 directory.
            Attempting to uninstall did_you_mean-1.3.0
            Unable to uninstall did_you_mean-1.3.0:
                Gem::FilePermissionError: You don't have write permissions \
                for the /Library/Ruby/Gems/2.6.0 directory.
            Clean up complete
        """
        super(Gem, self).cleanup()
        self.run_cli("cleanup", self.global_args)



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

""" This is the last place in which we support Python 2.7 to allow BitBar
plugin to be tested out in our CI workflow.
"""

try:
    # Python 3.
    from importlib.machinery import SourceFileLoader
    from pathlib import Path

    PY = 3
except ImportError:
    # Python 2.
    import imp
    from os import path

    PY = 2


# Manually import BitBar plugin source code as a module, because of its non
# Python-compliant filename with a double extension made the first dot
# interpreted as a submodule.
if PY == 2:
    here = path.dirname(path.abspath(__file__))
    bitbar_plugin = path.join(here, "meta_package_manager.7h.py")
    imp.load_source(__name__, bitbar_plugin)
else:
    SourceFileLoader(
        __name__,
        str(Path(__file__).parent.joinpath("meta_package_manager.7h.py").resolve()),
    ).load_module()



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

# pylint: disable=redefined-outer-name

import pytest

from .conftest import MANAGER_IDS
from .test_cli import CLISubCommandTests


@pytest.fixture
def subcmd():
    return "sync"


class TestSync(CLISubCommandTests):
    @pytest.mark.parametrize("mid", MANAGER_IDS)
    def test_single_manager(self, invoke, subcmd, mid):
        result = invoke("--manager", mid, subcmd)
        assert result.exit_code == 0
        self.check_manager_selection(result, {mid})



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import functools
import logging
import re
from datetime import datetime
from functools import partial
from io import TextIOWrapper
from operator import getitem, itemgetter
from pathlib import Path
from sys import __stdin__, __stdout__
from time import time as time_now

import click
import click_log
import tomlkit
from boltons.cacheutils import LRI, cached
from boltons.strutils import complement_int_list, int_ranges_from_int_list, strip_ansi
from cli_helpers.tabular_output import TabularOutputFormatter
from simplejson import dumps as json_dumps

from . import __version__, logger, CLI_NAME
from .base import CLI_FORMATS, CLIError, PackageManager
from .managers import pool
from .platform import CURRENT_OS_ID, WINDOWS, os_label
from .version import TokenizedString

# Initialize the table formatter.
table_formatter = TabularOutputFormatter()


# Register all rendering modes for table data.
RENDERING_MODES = {"json"}
RENDERING_MODES.update(table_formatter.supported_formats)
RENDERING_MODES = frozenset(RENDERING_MODES)

# List of unicode rendering modes that will fall back to ascii on windows.
# Windows has some hard time printing unicode characters to console output. It
# seems to be an effect of cp1252 encoding and/or click not able to transcode
# chars. Here is the traceback:
#
# File "(...)\meta_package_manager\cli.py", line 133, in print_table
#   click.echo(line)
# File "(...)\site-packages\click\utils.py", line 272, in echo
#   file.write(message)
# File "(...)\lib\encodings\cp1252.py", line 19, in encode
#   return codecs.charmap_encode(input,self.errors,encoding_table)[0]
# UnicodeEncodeError: 'charmap' codec can't encode characters in position
#   0-140: character maps to <undefined>
#
# Fortunately, I found the fundamental issue and I no longer need to blacklist
# some rendering modes. See: a3008f8c3a42efedd88378f087202b73d907bbb7 . I'll
# still keep the construct around just in case I need to quickly blacklist
# some.
WINDOWS_MODE_BLACKLIST = frozenset([])

# List of fields IDs allowed to be sorted.
SORTABLE_FIELDS = {
    "manager_id",
    "manager_name",
    "package_id",
    "package_name",
    "version",
}

# Pre-rendered UI-elements.
OK = click.style("✓", fg="green")
KO = click.style("✘", fg="red")


click_log.basic_config(logger)


def json(data):
    """Utility function to render data structure into pretty printed JSON.

    Also care of internal objects like `TokenizedString` and `Path`:
    """

    def serialize_objects(obj):
        if isinstance(obj, (TokenizedString, Path)):
            return str(obj)
        raise TypeError(repr(obj) + " is not JSON serializable.")

    return json_dumps(
        data,
        sort_keys=True,
        indent=4,
        separators=(",", ": "),
        default=serialize_objects,
    )


def print_table(header_defs, rows, sort_key=None):
    """ Utility to print a table and sort its content. """
    # Do not print anything, not even table headers if no rows.
    if not rows:
        return

    header_labels = [label for label, _ in header_defs]

    # Check there is no duplicate column IDs.
    header_ids = [col_id for _, col_id in header_defs if col_id]
    assert len(header_ids) == len(set(header_ids))

    # Default sorting follows the order of headers.
    sort_order = list(range(len(header_defs)))

    # Move the sorting key's index in the front of priority.
    if sort_key and sort_key in header_ids:
        # Build an index of column id's position.
        col_index = {col_id: i for i, (_, col_id) in enumerate(header_defs) if col_id}
        sort_column_index = col_index[sort_key]
        sort_order.remove(sort_column_index)
        sort_order.insert(0, sort_column_index)

    def sort_method(line):
        """Serialize line's content for natural sorting.

        1. Extract each cell value in the order provided by `sort_order`;
        2. Strip terminal color formating;
        3. Then tokenize each cell's content for user-friendly natural sorting.
        """
        sorting_key = []
        for cell in itemgetter(*sort_order)(line):
            if isinstance(cell, TokenizedString):
                key = cell
            else:
                key = TokenizedString(strip_ansi(cell))
            sorting_key.append(key)
        return tuple(sorting_key)

    for line in table_formatter.format_output(
        sorted(rows, key=sort_method), header_labels, disable_numparse=True
    ):
        click.echo(line.encode("utf-8"))


def print_stats(data):
    """ Print statistics. """
    manager_stats = {infos["id"]: len(infos["packages"]) for infos in data.values()}
    total_installed = sum(manager_stats.values())
    per_manager_totals = ", ".join(
        [
            "{}: {}".format(k, v)
            for k, v in sorted(manager_stats.items(), key=itemgetter(1), reverse=True)
        ]
    )
    if per_manager_totals:
        per_manager_totals = " ({})".format(per_manager_totals)
    plural = "s" if total_installed > 1 else ""
    click.echo(f"{total_installed} package{plural} total{per_manager_totals}.")


class timeit:
    """Decorator to measure and print elapsed execution time of a function."""

    def __call__(self, func):
        @functools.wraps(func)
        def decorated(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return decorated

    def __enter__(self):
        self.measure_time = click.get_current_context().obj["time"]
        if self.measure_time:
            self.start_time = time_now()

    def __exit__(self, *args, **kwargs):
        if self.measure_time:
            elapsed = time_now() - self.start_time
            click.echo(f"Execution time: {elapsed:.3} seconds.")


@click.group()
@click_log.simple_verbosity_option(
    logger,
    default="INFO",
    metavar="LEVEL",
    help="Either CRITICAL, ERROR, WARNING, INFO or DEBUG. Defaults to INFO.",
)
@click.option(
    "-m",
    "--manager",
    type=click.Choice(pool()),
    multiple=True,
    help="Restrict sub-command to a subset of package managers. Repeat to "
    "select multiple managers. Defaults to all.",
)
@click.option(
    "-e",
    "--exclude",
    type=click.Choice(pool()),
    multiple=True,
    help="Exclude a package manager. Repeat to exclude multiple managers. "
    "Defaults to none.",
)
@click.option(
    "--ignore-auto-updates/--include-auto-updates",
    default=True,
    help="Report all outdated packages, including those tagged as "
    "auto-updating. Defaults to include all packages. Only applies to "
    "'outdated' and 'upgrade' commands.",
)
@click.option(
    "-o",
    "--output-format",
    type=click.Choice(sorted(RENDERING_MODES)),
    default="fancy_grid",
    help="Rendering mode of the output. Defaults to fancy-grid.",
)
@click.option(
    "-s",
    "--sort-by",
    type=click.Choice(SORTABLE_FIELDS),
    default="manager_id",
    help="Sort results. Defaults to manager_id.",
)
@click.option(
    "--stats/--no-stats",
    default=True,
    help="Print per-manager package statistics. Active by default.",
)
@click.option(
    "--time/--no-time",
    default=False,
    help="Measure and print elapsed execution time. Inactive by default.",
)
@click.option(
    "--stop-on-error/--continue-on-error",
    default=True,
    help="Stop right "
    "away or continue operations on manager CLI error. Defaults to stop.",
)
@click.version_option(__version__)
@click.pass_context
def cli(
    ctx,
    manager,
    exclude,
    ignore_auto_updates,
    output_format,
    sort_by,
    stats,
    time,
    stop_on_error,
):
    """ CLI for multi-package manager upgrades. """
    # TODO: merge CLI parameters and config file here.
    # See: https://github.com/kdeldycke/meta-package-manager/issues/66

    level = logger.level
    level_name = logging._levelToName.get(level, level)
    logger.debug(f"Verbosity set to {level_name}.")

    # Target all available managers by default.
    target_ids = set(pool())
    # Only keeps the subset of selected by the user.
    if manager:
        target_ids = target_ids.intersection(manager)
    # Remove managers excluded by the user.
    target_ids = target_ids.difference(exclude)
    target_managers = [pool()[mid] for mid in sorted(target_ids)]

    # Apply manager-level options.
    for m_obj in target_managers:
        # Does the manager should raise on error or not.
        m_obj.raise_on_error = stop_on_error
        # Should we include auto-update packages or not?
        m_obj.ignore_auto_updates = ignore_auto_updates

    # Pre-filters inactive managers.
    def keep_available(manager):
        if manager.available:
            return True
        logger.warning(f"Skip unavailable {manager.id} manager.")

    # Use an iterator to not trigger log messages for the 'managers' subcommand
    # which is not using this variable.
    active_managers = filter(keep_available, target_managers)

    # Silence all log message for JSON rendering unless in debug mode.
    if output_format == "json" and level_name != "DEBUG":
        logger.setLevel(logging.CRITICAL * 2)

    # Setup the table formatter.
    if output_format != "json":

        # Fallback unicode-rendering to safe ascii on Windows.
        if CURRENT_OS_ID == WINDOWS and output_format in WINDOWS_MODE_BLACKLIST:
            output_format = "ascii"

        table_formatter.format_name = output_format

    # Load up global options to the context.
    ctx.obj = {
        "target_managers": target_managers,
        "active_managers": active_managers,
        "output_format": output_format,
        "sort_by": sort_by,
        "stats": stats,
        "time": time,
    }


@cli.command(short_help="List supported package managers and their location.")
@click.pass_context
@timeit()
def managers(ctx):
    """List all supported package managers and their presence on the system."""
    target_managers = ctx.obj["target_managers"]
    output_format = ctx.obj["output_format"]
    sort_by = ctx.obj["sort_by"]

    # Machine-friendly data rendering.
    if output_format == "json":
        manager_data = {}
        # Build up the data structure of manager metadata.
        fields = [
            "name",
            "id",
            "supported",
            "cli_path",
            "executable",
            "version",
            "fresh",
            "available",
        ]
        for manager in target_managers:
            manager_data[manager.id] = {fid: getattr(manager, fid) for fid in fields}
            # Serialize errors at the last minute to gather all we encountered.
            manager_data[manager.id]["errors"] = list(
                {expt.error for expt in manager.cli_errors}
            )

        click.echo(json(manager_data))
        return

    # Human-friendly content rendering.
    table = []
    for manager in target_managers:

        # Build up the OS column content.
        os_infos = OK if manager.supported else KO
        if not manager.supported:
            os_infos += "  {} only".format(
                ", ".join(sorted([os_label(os_id) for os_id in manager.platforms]))
            )

        # Build up the CLI path column content.
        cli_infos = "{}  {}".format(
            OK if manager.cli_path else KO,
            manager.cli_path
            if manager.cli_path
            else "{!r} not found".format(manager.cli_name),
        )

        # Build up the version column content.
        version_infos = ""
        if manager.executable:
            version_infos = OK if manager.fresh else KO
            if manager.version:
                version_infos += "  {}".format(manager.version)
                if not manager.fresh:
                    version_infos += " {}".format(manager.requirement)

        table.append(
            [
                manager.name,
                click.style(manager.id, fg="green" if manager.fresh else "red"),
                os_infos,
                cli_infos,
                OK if manager.executable else "",
                version_infos,
            ]
        )

    print_table(
        [
            ("Package manager", "manager_name"),
            ("ID", "manager_id"),
            ("Supported", None),
            ("CLI", None),
            ("Executable", None),
            ("Version", "version"),
        ],
        table,
        sort_by,
    )


@cli.command(short_help="Sync local package info.")
@click.pass_context
@timeit()
def sync(ctx):
    """ Sync local package metadata and info from external sources. """
    active_managers = ctx.obj["active_managers"]

    for manager in active_managers:
        manager.sync()


@cli.command(short_help="Cleanup local data.")
@click.pass_context
@timeit()
def cleanup(ctx):
    """ Cleanup local data and temporary artifacts. """
    active_managers = ctx.obj["active_managers"]

    for manager in active_managers:
        manager.cleanup()


@cli.command(short_help="List installed packages.")
@click.pass_context
@timeit()
def installed(ctx):
    """ List all packages installed on the system from all managers. """
    active_managers = ctx.obj["active_managers"]
    output_format = ctx.obj["output_format"]
    sort_by = ctx.obj["sort_by"]
    stats = ctx.obj["stats"]

    # Build-up a global dict of installed packages per manager.
    installed_data = {}

    for manager in active_managers:
        installed_data[manager.id] = {
            "id": manager.id,
            "name": manager.name,
            "packages": list(manager.installed.values()),
        }

        # Serialize errors at the last minute to gather all we encountered.
        installed_data[manager.id]["errors"] = list(
            {expt.error for expt in manager.cli_errors}
        )

    # Machine-friendly data rendering.
    if output_format == "json":
        click.echo(json(installed_data))
        return

    # Human-friendly content rendering.
    table = []
    for manager_id, installed_pkg in installed_data.items():
        table += [
            [
                info["name"],
                info["id"],
                manager_id,
                info["installed_version"] if info["installed_version"] else "?",
            ]
            for info in installed_pkg["packages"]
        ]

    # Sort and print table.
    print_table(
        [
            ("Package name", "package_name"),
            ("ID", "package_id"),
            ("Manager", "manager_id"),
            ("Installed version", "version"),
        ],
        table,
        sort_by,
    )

    if stats:
        print_stats(installed_data)


@cli.command(short_help="Search packages.")
@click.option(
    "--extended/--package-name",
    default=False,
    help="Extend search to additional package metadata like description, "
    "instead of restricting it package ID and name. Defaults to package ID "
    "search.",
)
@click.option(
    "--exact/--fuzzy",
    default=False,
    help="Only returns exact matches, or enable fuzzy search in substrings. "
    "Fuzzy by default.",
)
@click.argument("query", type=click.STRING, required=True)
@click.pass_context
@timeit()
def search(ctx, extended, exact, query):
    """ Search packages from all managers. """
    active_managers = ctx.obj["active_managers"]
    output_format = ctx.obj["output_format"]
    sort_by = ctx.obj["sort_by"]
    stats = ctx.obj["stats"]

    # Build-up a global list of package matches per manager.
    matches = {}

    for manager in active_managers:
        matches[manager.id] = {
            "id": manager.id,
            "name": manager.name,
            "packages": list(manager.search(query, extended, exact).values()),
        }

        # Serialize errors at the last minute to gather all we encountered.
        matches[manager.id]["errors"] = list(
            {expt.error for expt in manager.cli_errors}
        )

    # Machine-friendly data rendering.
    if output_format == "json":
        click.echo(json(matches))
        return

    # Prepare highlighting helpers.
    query_parts = {query}.union(map(str, TokenizedString(query)))

    @cached(LRI(max_size=1000))
    def highlight(string):
        # Ranges of character indices flagged for highlighting.
        ranges = set()

        # TODO: Fix upper-case matching, as tokenizer lower them down.

        for part in query_parts:
            # Search for occurrences of query parts in original string.
            if part in string:
                # Flag matching substrings for highlighting.
                occurrences = [match.start() for match in re.finditer(part, string)]

                for match_start in occurrences:
                    match_end = match_start + len(part) - 1
                    ranges.add("{}-{}".format(match_start, match_end))

        # Reduce index ranges, compute complement ranges, transform them to
        # list of integers.
        ranges = ",".join(ranges)
        bold_ranges = int_ranges_from_int_list(ranges)
        normal_ranges = int_ranges_from_int_list(
            complement_int_list(ranges, range_end=len(string))
        )

        # Apply style to range of characters flagged as matching.
        styled_str = ""
        for i, j in sorted(bold_ranges + normal_ranges):
            segment = getitem(string, slice(i, j + 1))
            if (i, j) in bold_ranges:
                segment = click.style(segment, bold=True)
            styled_str += segment

        return styled_str

    # Human-friendly content rendering.
    table = []
    for manager_id, matching_pkg in matches.items():
        table += [
            [
                highlight(info["name"]),
                highlight(info["id"]),
                manager_id,
                info["latest_version"] if info["latest_version"] else "?",
            ]
            for info in matching_pkg["packages"]
        ]

    # Sort and print table.
    print_table(
        [
            ("Package name", "package_name"),
            ("ID", "package_id"),
            ("Manager", "manager_id"),
            ("Latest version", "version"),
        ],
        table,
        sort_by,
    )

    if stats:
        print_stats(matches)


@cli.command(short_help="List outdated packages.")
@click.option(
    "-c",
    "--cli-format",
    type=click.Choice(CLI_FORMATS),
    default="plain",
    help="Format of CLI fields in JSON output. Defaults to plain.",
)
@click.pass_context
@timeit()
def outdated(ctx, cli_format):
    """List available package upgrades and their versions for each manager."""
    active_managers = ctx.obj["active_managers"]
    output_format = ctx.obj["output_format"]
    sort_by = ctx.obj["sort_by"]
    stats = ctx.obj["stats"]

    render_cli = partial(PackageManager.render_cli, cli_format=cli_format)

    # Build-up a global list of outdated packages per manager.
    outdated_data = {}

    for manager in active_managers:

        packages = list(map(dict, manager.outdated.values()))
        for info in packages:
            info.update({"upgrade_cli": render_cli(manager.upgrade_cli(info["id"]))})

        outdated_data[manager.id] = {
            "id": manager.id,
            "name": manager.name,
            "packages": packages,
        }

        # Do not include the full-upgrade CLI if we did not detect any outdated
        # package.
        if manager.outdated:
            try:
                upgrade_all_cli = manager.upgrade_all_cli()
            except NotImplementedError:
                # Fallback on mpm itself which is capable of simulating a full
                # upgrade.
                upgrade_all_cli = [CLI_NAME, "--manager", manager.id, "upgrade"]
            outdated_data[manager.id]["upgrade_all_cli"] = render_cli(upgrade_all_cli)

        # Serialize errors at the last minute to gather all we encountered.
        outdated_data[manager.id]["errors"] = list(
            {expt.error for expt in manager.cli_errors}
        )

    # Machine-friendly data rendering.
    if output_format == "json":
        click.echo(json(outdated_data))
        return

    # Human-friendly content rendering.
    table = []
    for manager_id, outdated_pkg in outdated_data.items():
        table += [
            [
                info["name"],
                info["id"],
                manager_id,
                info["installed_version"] if info["installed_version"] else "?",
                info["latest_version"],
            ]
            for info in outdated_pkg["packages"]
        ]

    # Sort and print table.
    print_table(
        [
            ("Package name", "package_name"),
            ("ID", "package_id"),
            ("Manager", "manager_id"),
            ("Installed version", "version"),
            ("Latest version", None),
        ],
        table,
        sort_by,
    )

    if stats:
        print_stats(outdated_data)


@cli.command(short_help="Upgrade all packages.")
@click.option(
    "-d",
    "--dry-run",
    is_flag=True,
    default=False,
    help="Do not actually perform any upgrade, just simulate CLI calls.",
)
@click.pass_context
@timeit()
def upgrade(ctx, dry_run):
    """ Perform a full package upgrade on all available managers. """
    active_managers = ctx.obj["active_managers"]

    for manager in active_managers:

        logger.info(f"Updating all outdated packages from {manager.id}...")

        try:
            output = manager.upgrade_all(dry_run=dry_run)
        except CLIError as expt:
            logger.error(expt.error)

        if output:
            logger.info(output)


@cli.command(short_help="Save installed packages to a TOML file.")
@click.argument("toml_output", type=click.File("w"), default="-")
@click.pass_context
@timeit()
def backup(ctx, toml_output):
    """Dump the list of installed packages to a TOML file.

    By default the generated TOML content is displayed directly in the console
    output. So `mpm backup` is the same as a call to `mpm backup -`. To have
    the result written in a file on disk, specify the output file like so:
    `mpm backup ./mpm-packages.toml`.

    The TOML file can then be safely consumed by the `mpm restore` command.
    """
    active_managers = ctx.obj["active_managers"]
    stats = ctx.obj["stats"]

    # XXX Hack for unittests to pass, while we wait for
    # https://github.com/pallets/click/pull/1497
    if isinstance(toml_output, TextIOWrapper):
        toml_output = __stdout__

    is_stdout = toml_output is __stdout__
    toml_filepath = toml_output.name if is_stdout else Path(toml_output.name).resolve()
    logger.info(f"Backup package list to {toml_filepath}")

    if not is_stdout:
        if toml_filepath.exists() and not toml_filepath.is_file():
            logger.error("Target file exist and is not a file.")
            return
        if toml_filepath.suffix.lower() != ".toml":
            logger.error("Target file is not a TOML file.")
            return

    # Initialize the TOML structure.
    doc = tomlkit.document()
    # Leave some metadata as comment.
    doc.add(tomlkit.comment("Generated by {} {}.".format(CLI_NAME, __version__)))
    doc.add(tomlkit.comment("Timestamp: {}.".format(datetime.now().isoformat())))

    installed_data = {}

    # Create one section for each package manager.
    for manager in active_managers:
        logger.info(f"Dumping packages from {manager.id}...")

        # Prepare data for stats.
        installed_data[manager.id] = {
            "id": manager.id,
            "packages": manager.installed.values(),
        }

        manager_section = tomlkit.table()
        pkg_data = sorted(
            [(p["id"], p["installed_version"]) for p in manager.installed.values()]
        )
        for package_id, package_version in pkg_data:
            # Version specifier is inspired by Poetry.
            manager_section.add(package_id, "^{}".format(package_version))
        if pkg_data:
            doc.add(manager.id, manager_section)

    toml_output.write(tomlkit.dumps(doc, sort_keys=True))

    if stats:
        print_stats(installed_data)


@cli.command(short_help="Install packages in batch as specified by TOML files.")
@click.argument("toml_files", type=click.File("r"), required=True, nargs=-1)
@click.pass_context
@timeit()
def restore(ctx, toml_files):
    """Read TOML files then install or upgrade each package referenced in
    them.
    """
    active_managers = ctx.obj["active_managers"]

    for toml_input in toml_files:

        toml_filepath = (
            toml_input.name
            if toml_input is __stdin__
            else Path(toml_input.name).resolve()
        )
        logger.info(f"Load package list from {toml_filepath}")

        doc = tomlkit.parse(toml_input.read())

        # List unrecognized sections.
        ignored_sections = [
            "[{}]".format(section) for section in doc if section not in pool()
        ]
        if ignored_sections:
            plural = "s" if len(ignored_sections) > 1 else ""
            sections = ", ".join(ignored_sections)
            logger.warning(f"Ignore {sections} section{plural}.")

        for manager in active_managers:
            if manager.id not in doc:
                logger.warning(f"No [{manager.id}] section found.")
                continue
            logger.info(f"Restore {manager.id} packages...")
            logger.warning("Installation of packages not implemented yet.")
            # for package_id, version in doc[manager.id].items():
            #    raise NotImplemented



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# <bitbar.title>Meta Package Manager</bitbar.title>
# <bitbar.version>v3.4.2</bitbar.version>
# <bitbar.author>Kevin Deldycke</bitbar.author>
# <bitbar.author.github>kdeldycke</bitbar.author.github>
# <bitbar.desc>List outdated packages and manage upgrades.</bitbar.desc>
# <bitbar.dependencies>python,mpm</bitbar.dependencies>
# <bitbar.image>https://i.imgur.com/CiQpQ42.png</bitbar.image>
# <bitbar.abouturl>https://github.com/kdeldycke/meta-package-manager</bitbar.abouturl>

"""
Bitbar plugin for Meta Package Manager (a.k.a. the :command:`mpm` CLI).

Default update cycle is set to 7 hours so we have a chance to get user's
attention once a day. Higher frequency might ruin the system as all checks are
quite resource intensive, and Homebrew might hit GitHub's API calls quota.

Minimal BitBar requirement is Mac OS X Lion (10.7), which ships with Python
2.7.1. So this plugin is supposed to support Python 2.7.1 or newer.
"""

from __future__ import print_function, unicode_literals

import json
import os
import sys
from operator import itemgetter
from subprocess import PIPE, Popen

PY2 = sys.version_info[0] == 2


SUBMENU_LAYOUT = bool(
    os.environ.get("BITBAR_MPM_SUBMENU", False)
    in {True, 1, "True", "true", "1", "y", "yes", "Yes"}
)
""" Define the rendering mode of outdated packages list.

Edit this script to force this constant to ``True``, or use the
``BITBAR_MPM_SUBMENU`` environment variable. This will replace the default flat
layout with an alternative structure where all upgrade actions are put into
submenus, one for each manager.
"""


# Make it easier to change font, sizes and colors of the output
# See https://github.com/matryer/bitbar#writing-plugins for details
# An alternate "good looking" font is "font=NotoMono size=13" (not installed
# on MacOS by default though) that matches the system font quite well.
FONTS = {
    "normal": "",  # Use default system font
    "summary": "",  # Package summary
    "package": "",  # Indiviual packages
    "error": "color=red font=Menlo size=12",  # Errors
}
# Use a monospaced font when using submenus.
if SUBMENU_LAYOUT:
    FONTS["summary"] = "font=Menlo size=12"


def fix_environment():
    """Tweak environment variable to find non-default system-wide binaries.

    macOS does not put ``/usr/local/bin`` or ``/opt/local/bin`` in the ``PATH``
    for GUI apps. For some package managers this is a problem. Additioanlly
    Homebrew and Macports are using different pathes. So, to make sure we can
    always get to the necessary binaries, we overload the path. Current
    preference order would equate to Homebrew, Macports, then system.
    """
    os.environ["PATH"] = ":".join(
        [
            "/usr/local/bin",
            "/usr/local/sbin",
            "/opt/local/bin",
            "/opt/local/sbin",
            os.environ.get("PATH", ""),
        ]
    )

    # Python 3 Surrogate Handling. See:
    # https://click.pocoo.org/6/python3/#python-3-surrogate-handling
    os.environ["LC_ALL"] = os.environ["LANG"] = "en_US.UTF-8"


def run(*args):
    """Run a shell command, return error code, output and error message."""
    assert isinstance(args, tuple)
    try:
        process = Popen(args, stdout=PIPE, stderr=PIPE)
    except OSError:
        return None, None, "`{}` executable not found.".format(args[0])
    output, error = process.communicate()
    return (
        process.returncode,
        output.decode("utf-8") if output else None,
        error.decode("utf-8") if error else None,
    )


def echo(message):
    """Print message to the output.

    Not unlike ``click.echo()``, this method is required to support
    discrepencies in the way strings are handled in different Python versions
    and platforms.
    """
    if PY2:
        message = message.encode("utf-8")
    print(message)


def print_error_header():
    """Generic header for blockng error."""
    echo("❌ | dropdown=false")
    echo("---")


def print_error(message, submenu=""):
    """Print a formatted error line by line.

    A red, fixed-width font is used to preserve traceback and exception layout.
    """
    for line in message.strip().splitlines():
        echo(
            "{}{} | {f_error} trim=false emojize=false"
            "".format(submenu, line, f_error=FONTS["error"])
        )


def print_cli_item(item):
    """Print two CLI entries:
    * one that is silent
    * a second one that is the exact copy of the above but forces the execution
      by the way of a visible terminal
    """
    echo("{} terminal=false".format(item))
    echo("{} terminal=true alternate=true".format(item))


def print_package_items(packages, submenu=""):
    """Print a menu entry for each outdated packages available for upgrade."""
    for pkg_info in packages:
        print_cli_item(
            "{}{name} {installed_version} → {latest_version} | {upgrade_cli}"
            " refresh=true {f_package} emojize=false".format(
                submenu, f_package=FONTS["package"], **pkg_info
            )
        )


def print_upgrade_all_item(manager, submenu=""):
    """Print the menu entry to upgrade all outdated package of a manager."""
    if manager.get("upgrade_all_cli"):
        if SUBMENU_LAYOUT:
            echo("-----")
        print_cli_item(
            "{}Upgrade all | {} refresh=true {f_normal}".format(
                submenu, manager["upgrade_all_cli"], f_normal=FONTS["normal"]
            )
        )


def print_menu():
    """Print menu structure using BitBar's plugin API.

    See: https://github.com/matryer/bitbar#plugin-api
    """
    # Search for generic mpm CLI on system.
    code, _, error = run("mpm")
    # mpm CLI hasn't been found on the system. Propose to the user to install
    # or upgrade it.
    if code or error:
        print_error_header()
        print_error(error)
        echo("---")
        echo(
            "Install / upgrade `mpm` CLI. | bash=python param1=-m param2=pip "
            "param3=install param4=--upgrade "
            'param5=\\"meta-package-manager>=2.7.0\\" terminal=true '
            "refresh=true {f_error}".format(f_error=FONTS["error"])
        )
        return

    # Force a sync of all local package databases.
    run("mpm", "sync")

    # Fetch outdated package form all package manager available on the system.
    _, output, error = run(
        "mpm", "--output-format", "json", "outdated", "--cli-format", "bitbar"
    )

    # Bail-out immediately on errors related to mpm self-execution or if mpm is
    # not able to produce any output.
    if error or not output:
        print_error_header()
        print_error(error)
        return

    # Sort outdated packages by manager's name.
    managers = sorted(json.loads(output).values(), key=itemgetter("name"))

    # Print menu bar icon with number of available upgrades.
    total_outdated = sum([len(m["packages"]) for m in managers])
    total_errors = sum([len(m.get("errors", [])) for m in managers])
    echo(
        "↑{}{} | dropdown=false".format(
            total_outdated, " ⚠️{}".format(total_errors) if total_errors else ""
        )
    )

    # Print a full detailed section for each manager.
    submenu = "--" if SUBMENU_LAYOUT else ""

    if SUBMENU_LAYOUT:
        # Compute maximal manager's name length.
        label_max_length = max([len(m["name"]) for m in managers])
        max_outdated = max([len(m["packages"]) for m in managers])
        echo("---")

    for manager in managers:
        package_label = "package{}".format("s" if len(manager["packages"]) > 1 else "")

        if SUBMENU_LAYOUT:
            # Non-flat layout use a compact table-like rendering of manager
            # summary.
            echo(
                "{error}{0:<{max_length}} {1:>{max_outdated}} {2:<8} | "
                "{f_summary} emojize=false".format(
                    manager["name"] + ":",
                    len(manager["packages"]),
                    package_label,
                    error="⚠️ " if manager.get("errors", None) else "",
                    max_length=label_max_length + 1,
                    max_outdated=len(str(max_outdated)),
                    f_summary=FONTS["summary"],
                )
            )
        else:
            echo("---")
            echo(
                "{0} outdated {1} {2} | {f_summary} emojize=false".format(
                    len(manager["packages"]),
                    manager["name"],
                    package_label,
                    f_summary=FONTS["summary"],
                )
            )

        print_package_items(manager["packages"], submenu)

        print_upgrade_all_item(manager, submenu)

        for error_msg in manager.get("errors", []):
            echo("-----" if SUBMENU_LAYOUT else "---")
            print_error(error_msg, submenu)


if __name__ == "__main__":
    fix_environment()
    print_menu()



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from types import FunctionType

import pytest

from ..managers import pool
from ..platform import (
    ALL_OS_LABELS,
    CURRENT_OS_ID,
    CURRENT_OS_LABEL,
    LINUX,
    MACOS,
    OS_DEFINITIONS,
    WINDOWS,
    current_os,
    is_linux,
    is_macos,
    is_windows,
    os_label,
)
from .conftest import MANAGER_IDS, unless_linux, unless_macos, unless_windows


def test_mutual_exclusion():
    if is_linux():
        assert not is_macos()
        assert not is_windows()
        assert CURRENT_OS_ID == LINUX
        assert CURRENT_OS_LABEL == os_label(LINUX)
    if is_macos():
        assert not is_linux()
        assert not is_windows()
        assert CURRENT_OS_ID == MACOS
        assert CURRENT_OS_LABEL == os_label(MACOS)
    if is_windows():
        assert not is_linux()
        assert not is_macos()
        assert CURRENT_OS_ID == WINDOWS
        assert CURRENT_OS_LABEL == os_label(WINDOWS)


def test_os_definitions():
    assert isinstance(OS_DEFINITIONS, dict)
    # Each OS definition must be unique.
    assert isinstance(ALL_OS_LABELS, frozenset)
    assert len(OS_DEFINITIONS) == len(ALL_OS_LABELS)
    for os_id, data in OS_DEFINITIONS.items():
        # OS ID.
        assert isinstance(os_id, str)
        assert os_id
        assert os_id.isascii()
        assert os_id.isalpha()
        assert os_id.islower()
        # Metadata.
        assert isinstance(data, tuple)
        assert len(data) == 2
        label, os_flag = data
        # OS label.
        assert label
        assert isinstance(label, str)
        assert label.isascii()
        assert label.isalpha()
        assert label in ALL_OS_LABELS
        # OS identification function.
        assert isinstance(os_flag, bool)
        os_id_func_name = "is_{}".format(os_id)
        assert os_id_func_name in globals()
        os_id_func = globals()[os_id_func_name]
        assert isinstance(os_id_func, FunctionType)
        assert isinstance(os_id_func(), bool)
        assert os_id_func() == os_flag


def test_current_os_func():
    # Function.
    os_id, label = current_os()
    assert os_id in OS_DEFINITIONS
    assert label in [os[0] for os in OS_DEFINITIONS.values()]
    # Constants.
    assert os_id == CURRENT_OS_ID
    assert label == CURRENT_OS_LABEL


def test_os_label():
    os_id, os_name = current_os()
    assert os_label(os_id) == os_name


def test_blacklisted_manager():
    """ Check all managers are accounted for on each platforms. """
    blacklists = {
        LINUX: {"brew", "cask", "mas"},
        MACOS: {"apt", "flatpak", "opkg", "snap"},
        WINDOWS: {"apt", "cask", "brew", "flatpak", "mas", "opkg", "snap"},
    }
    blacklist = blacklists[current_os()[0]]
    # List of supported managers on the current platform.
    supported = {m.id for m in pool().values() if m.supported}
    assert supported == MANAGER_IDS - blacklist


# Test unittest decorator helpers.


@unless_linux
def test_unless_linux():
    assert is_linux()
    assert not is_macos()
    assert not is_windows()


@unless_macos
def test_unless_macos():
    assert not is_linux()
    assert is_macos()
    assert not is_windows()


@unless_windows
def test_unless_windows():
    assert not is_linux()
    assert not is_macos()
    assert is_windows()



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import re
import sys
import unittest
from collections import Counter

from .. import bitbar


"""
Like BitBar, this plugin is supposed to run smoothly with Python 2.7.1 or
newer.
"""


@unittest.skipUnless(sys.platform == "darwin", "macOS required")
class TestBibarPlugin(unittest.TestCase):
    """This is the only test suite that is still using unittest module instead
    of pytest.
    """

    common_checklist = [
        # Menubar line. Required.
        (r"↑\d+ (⚠️\d+ )?\| dropdown=false$", True),
        # Package upgrade line. Optional: there might be no package to
        # upgrade.
        (r"(--)?\S+ \S+ → \S+ \| bash=.+$", False),
        # Submenu marker line. Required.
        (r"-{3,5}$", True),
        # Upgrade all line. Required.
        (r"(--)?Upgrade all \| bash=.+$", False),
        # Error line. Optional.
        (r"(--)?.+ \| color=red font=Menlo size=12 trim=false emojize=false$", False),
    ]

    def bitbar_output_checks(self, checklist, env=None):
        if env:
            for var, value in env.items():
                os.environ[var] = value
        code, output, error = bitbar.run(bitbar.__file__)
        if env:
            for var in env:
                del os.environ[var]
        self.assertEqual(code, 0)
        self.assertIsNone(error)

        checks = checklist + self.common_checklist

        match_counter = Counter()

        for line in output.splitlines():
            # The line is expected to match at least one regexp.
            matches = False
            for index, (regex, _) in enumerate(checks):
                if re.match(regex, line):
                    matches = True
                    match_counter[index] += 1
                    break
            if not matches:
                self.fail(
                    "BitBar output line {!r} did not match any regexp.".format(line)
                )

        # Check all required regexp did match at least once.
        for index, (regex, required) in enumerate(checks):
            if required and not match_counter[index]:
                self.fail(
                    "{!r} regex did not match any BitBar plugin output line."
                    "".format(regex)
                )

    def test_simple_call(self):
        """ Check default rendering is flat: no submenu. """
        self.bitbar_output_checks(
            [
                # Summary package statistics. Required.
                (r"\d+ outdated .+ packages? \|  emojize=false$", True),
            ]
        )

    def test_submenu_rendering(self):
        self.bitbar_output_checks(
            [
                # Submenu entry line with summary. Required.
                (r".+:\s+\d+ package(s| ) \| font=Menlo size=12 emojize=false$", True),
            ],
            env={"BITBAR_MPM_SUBMENU": "True"},
        )



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

# pylint: disable=redefined-outer-name

import os
from textwrap import indent

import click
import pytest
from boltons.iterutils import flatten, same
from boltons.strutils import strip_ansi
from boltons.tbutils import ExceptionInfo
from click.testing import CliRunner

from ..bitbar import run as bitbar_run
from ..cli import cli
from .. import CLI_NAME
from ..platform import is_linux, is_macos, is_windows

""" Fixtures, configuration and helpers for tests. """


MANAGER_IDS = frozenset(
    [
        "apm",
        "apt",
        "brew",
        "cask",
        "composer",
        "flatpak",
        "gem",
        "mas",
        "npm",
        "opkg",
        "pip",
        "snap",
        "yarn",
    ]
)
""" Hard-coded list of all supported manager IDs. """


DESTRUCTIVE_MODE = bool(
    os.environ.get("DESTRUCTIVE_TESTS", False) not in {True, 1, "True", "true", "1"}
)
""" Pre-computed boolean flag indicating if destructive mode is activated by
the presence of a ``DESTRUCTIVE_TESTS`` environment variable set to ``True``.
"""


destructive = pytest.mark.skipif(DESTRUCTIVE_MODE, reason="destructive test")
""" Pytest mark to skip a test unless destructive mode is allowed.

.. todo:

    Test destructive test assessment.
"""


non_destructive = pytest.mark.skipif(
    not DESTRUCTIVE_MODE, reason="non-destructive test"
)
""" Pytest mark to skip a test unless destructive mode is allowed.

.. todo:

    Test destructive test assessment.
"""


unless_linux = pytest.mark.skipif(not is_linux(), reason="Linux required")
""" Pytest mark to skip a test unless it is run on a Linux system. """


unless_macos = pytest.mark.skipif(not is_macos(), reason="macOS required")
""" Pytest mark to skip a test unless it is run on a macOS system. """


unless_windows = pytest.mark.skipif(not is_windows(), reason="Windows required")
""" Pytest mark to skip a test unless it is run on a Windows system. """


def print_cli_output(cmd, output):
    """ Simulate CLI output. Used to print debug traces in test results. """
    print("\n► {}".format(click.style(" ".join(cmd), fg="white")))
    if output:
        print(indent(output, "  "))


def run_cmd(*args):
    """Run a system command, print output and return results.

    Relies on robust ``run`` function implemented in BitBar plugin.
    """
    code, output, error = bitbar_run(*args)

    print_cli_output(args, output)

    # Print some more debug info.
    print(click.style("Return code: {}".format(code), fg="yellow"))
    if error:
        print(indent(click.style(error, fg="red"), "  "))

    return code, output, error


@pytest.fixture
def runner():
    runner = CliRunner(mix_stderr=False)
    with runner.isolated_filesystem():
        yield runner


@pytest.fixture
def invoke(runner):
    """ Executes Click's CLI, print output and return results. """

    # XXX Python 3-only function signature.
    # TODO: Use the commented signature below once we drop special Python 2.7
    # unittest for BitBar.
    # def _run(*args, color=False):
    def _run(*args, **kwargs):
        color = kwargs.get("color", False)

        # We allow for nested iterables and None values as args for
        # convenience. We just need to flatten and filters them out.
        args = list(filter(None.__ne__, flatten(args)))
        if args:
            assert same(map(type, args), str)

        result = runner.invoke(cli, args, color=color)

        # Strip colors out of results.
        result.stdout_bytes = strip_ansi(result.stdout_bytes)
        result.stderr_bytes = strip_ansi(result.stderr_bytes)

        print_cli_output([CLI_NAME] + args, result.output)

        # Print some more debug info.
        print(result)
        if result.exception:
            print(ExceptionInfo.from_exc_info(*result.exc_info).get_formatted())

        return result

    return _run


@pytest.fixture(scope="class")
def subcmd():
    """Fixture used in `test_cli_*.py` files to set the sub-command in all CLI
    calls.

    Must returns a string or an iterable of strings. Defaults to `None`, which
    allows tests relying on this fixture to selectively skip running.
    """
    return None



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import re

from ..base import PackageManager
from ..platform import MACOS
from ..version import parse_version
from . import logger


class MAS(PackageManager):

    platforms = frozenset([MACOS])

    # 'mas search' output has been fixed in 1.6.1:
    # https://github.com/mas-cli/mas/pull/205
    requirement = "1.6.1"

    name = "Mac AppStore"

    def get_version(self):
        """ Fetch version from ``mas version`` output."""
        return parse_version(self.run_cli("version"))

    @property
    def installed(self):
        """Fetch installed packages from ``mas list`` output.

        Raw CLI output samples:

        .. code-block:: shell-session

            ► mas list
            408981434 iMovie (10.1.4)
            747648890 Telegram (2.30)
        """
        installed = {}

        output = self.run_cli(self.global_args, "list")

        if output:
            regexp = re.compile(r"(\d+) (.*) \((\S+)\)$")
            for package in output.splitlines():
                match = regexp.match(package)
                if match:
                    package_id, package_name, installed_version = match.groups()
                    installed[package_id] = {
                        "id": package_id,
                        "name": package_name,
                        # Normalize unknown version. See:
                        # https://github.com/mas-cli/mas/commit
                        # /1859eaedf49f6a1ebefe8c8d71ec653732674341
                        "installed_version": parse_version(installed_version),
                    }

        return installed

    def search(self, query, extended, exact):
        """Fetch matching packages from ``mas search`` output.

        Raw CLI output samples:

        .. code-block:: shell-session

            ► mas search python
               689176796  Python Runner   (1.3)
               630736088  Learning Python (1.0)
               945397020  Run Python      (1.0)
              1164498373  PythonGames     (1.0)
              1400050251  Pythonic        (1.0.0)
        """
        matches = {}

        if extended:
            logger.warning(
                f"Extended search not supported for {self.id}. Fallback to Fuzzy."
            )

        output = self.run_cli(self.global_args, "search", query)

        if output:
            regexp = re.compile(r"(\d+) (.*)$")

            regexp = re.compile(
                r"""
                (?P<package_id>\d+)
                \s+
                (?P<package_name>.+?)
                \s+
                \(
                    (?P<version>\S+)
                \)
                """,
                re.MULTILINE | re.VERBOSE,
            )

            for package_id, package_name, version in regexp.findall(output):

                # Filters out fuzzy matches, only keep stricly matching
                # packages.
                if exact and query not in (package_id, package_name):
                    continue

                matches[package_id] = {
                    "id": package_id,
                    "name": package_name,
                    "latest_version": parse_version(version),
                }

        return matches

    @property
    def outdated(self):
        """Fetch outdated packages from ``mas outdated`` output.

        Raw CLI output samples:

        .. code-block:: shell-session

            ► mas outdated

        .. todo

            An example of ``mas outdated`` output is missing above.
        """
        outdated = {}

        output = self.run_cli(self.global_args, "outdated")

        if output:
            regexp = re.compile(r"(\d+) (.*) \((\S+) -> (\S+)\)$")
            for package in output.splitlines():
                match = regexp.match(package)
                if match:
                    (
                        package_id,
                        package_name,
                        installed_version,
                        latest_version,
                    ) = match.groups()
                    outdated[package_id] = {
                        "id": package_id,
                        "name": package_name,
                        "latest_version": parse_version(latest_version),
                        # Normalize unknown version. See:
                        # https://github.com/mas-cli/mas/commit
                        # /1859eaedf49f6a1ebefe8c8d71ec653732674341
                        "installed_version": parse_version(installed_version),
                    }

        return outdated

    def upgrade_cli(self, package_id=None):
        cmd = [self.cli_path, self.global_args, "upgrade"]
        if package_id:
            cmd.append(package_id)
        return cmd

    def upgrade_all_cli(self):
        return self.upgrade_cli()



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

""" Helpers and utilities to identify and handle platform idiosyncracies. """

import sys

from boltons.dictutils import FrozenDict

from . import logger

LINUX = "linux"
""" Constant used to identify OSes of the Linux family. """


MACOS = "macos"
""" Constant used to identify OSes of the macOS family. """


WINDOWS = "windows"
""" Constant used to identify OSes of the Windows family. """


def is_linux():
    """ Return `True` only if current platform is of the Linux family. """
    return sys.platform.startswith("linux")


def is_macos():
    """ Return `True` only if current platform is of the macOS family. """
    return sys.platform == "darwin"


def is_windows():
    """ Return `True` only if current platform is of the Windows family. """
    return sys.platform in ["win32", "cygwin"]


# Map OS IDs to evaluation function and OS labels.
OS_DEFINITIONS = FrozenDict(
    {
        LINUX: ("Linux", is_linux()),
        MACOS: ("macOS", is_macos()),
        WINDOWS: ("Windows", is_windows()),
    }
)


# Generare sets of recognized IDs and labels.
ALL_OS_LABELS = frozenset([label for label, _ in OS_DEFINITIONS.values()])


def os_label(os_id):
    """ Return platform label for user-friendly output. """
    return OS_DEFINITIONS[os_id][0]


logger.debug(f"Raw platform ID: {sys.platform}.")


def current_os():
    """ Return a 2-items `tuple` with ID and label of current OS. """
    for os_id, (os_name, os_flag) in OS_DEFINITIONS.items():
        if os_flag is True:
            return os_id, os_name
    raise SystemError("Unrecognized {} platform.".format(sys.platform))


CURRENT_OS_ID, CURRENT_OS_LABEL = current_os()



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import logging
import re
from pathlib import Path

import click
import pytest
import simplejson as json
from boltons.iterutils import flatten
from boltons.strutils import strip_ansi
from cli_helpers.tabular_output import TabularOutputFormatter

from .. import __version__, logger
from ..cli import RENDERING_MODES, WINDOWS_MODE_BLACKLIST
from .conftest import MANAGER_IDS, unless_windows

""" Common tests for all CLI basic features and templates for subcommands. """


def test_real_fs():
    """Check a simple test is not caught into the CLI runner fixture which is
    encapsulating all filesystem access into temporary directory structure."""
    assert str(Path(__file__)).startswith(str(Path.cwd()))


def test_temporary_fs(runner):
    """Check the CLI runner fixture properly encapsulated the filesystem in
    temporary directory."""
    assert not str(Path(__file__)).startswith(str(Path.cwd()))


class TestBaseCLI:

    """This collection is testing basic CLI behavior shared by all
    subcommands.

    Also regroups tests not involving subcommands.

    Also includes a bunch of tests performed once on an arbitrary sub-command,
    for situation when the tested behavior is shared by all subcommands. In
    that case we choosed `managers` as it is a safe read-only operation
    supposed to work on all platforms whatever the environment.
    """

    def test_bare_call(self, invoke):
        result = invoke()
        assert result.exit_code == 0
        assert "Usage: " in result.stdout
        assert not result.stderr

    def test_main_help(self, invoke):
        result = invoke("--help")
        assert result.exit_code == 0
        assert "Usage: " in result.stdout
        assert not result.stderr

    def test_version(self, invoke):
        result = invoke("--version")
        assert result.exit_code == 0
        assert __version__ in result.stdout
        assert not result.stderr

    def test_unknown_option(self, invoke):
        result = invoke("--blah")
        assert result.exit_code == 2
        assert not result.stdout
        assert "Error: no such option: --blah" in result.stderr

    def test_unknown_command(self, invoke):
        result = invoke("blah")
        assert result.exit_code == 2
        assert not result.stdout
        assert "Error: No such command 'blah'." in result.stderr

    def test_required_command(self, invoke):
        result = invoke("--verbosity", "DEBUG")
        assert result.exit_code == 2
        assert not result.stdout
        assert "Error: Missing command." in result.stderr

    def test_unrecognized_verbosity(self, invoke):
        result = invoke("--verbosity", "random", "managers")
        assert result.exit_code == 2
        assert not result.stdout
        assert "Error: Invalid value for '--verbosity' / '-v'" in result.stderr

    @pytest.mark.parametrize("level", ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"])
    def test_verbosity(self, invoke, level):
        result = invoke("--verbosity", level, "managers")
        assert result.exit_code == 0
        assert logger.level == getattr(logging, level)
        assert "══════" in result.stdout
        if level == "DEBUG":
            assert "debug: " in result.stderr
        else:
            assert "debug: " not in result.stderr

    @unless_windows
    @pytest.mark.parametrize("mode", WINDOWS_MODE_BLACKLIST)
    def test_check_failing_unicode_rendering(self, mode):
        """Check internal assumption that some rendering unicode table
        rendering modes fails in Windows console."""
        table_formatter = TabularOutputFormatter(mode)
        with pytest.raises(UnicodeEncodeError):
            click.echo(
                table_formatter.format_output(
                    ((1, 87), (2, 80), (3, 79)), ("day", "temperature")
                )
            )


class CLISubCommandTests:

    """All these tests runs on each subcommand.

    This class doesn't starts with `Test` as it is meant to be used as a
    template, inherited sub-command specific files.
    """

    def test_help(self, invoke, subcmd):
        result = invoke(subcmd, "--help")
        assert result.exit_code == 0
        assert "Usage: " in result.stdout
        assert flatten([subcmd])[0] in result.stdout
        assert not result.stderr

    @pytest.mark.parametrize("opt_stats", ["--stats", "--no-stats", None])
    @pytest.mark.parametrize("opt_timer", ["--time", "--no-time", None])
    def test_options(self, invoke, subcmd, opt_stats, opt_timer):
        """ Test the result on all combinations of optional options. """
        result = invoke(opt_stats, opt_timer, subcmd)
        assert result.exit_code == 0

    @staticmethod
    def check_manager_selection(result, selected=MANAGER_IDS):
        """Check inclusion and exclusion of a set of managers.

        Check all manager are there by default.

        .. todo:

            Parametrize/fixturize signals to pin point output depending on
            subcommand.
        """
        assert isinstance(selected, (frozenset, set))

        found_managers = set()
        skipped_managers = set()

        for mid in MANAGER_IDS:

            # List of signals indicating a package manager has been retained by
            # the CLI. Roughly sorted from most specific to more forgiving.
            signals = [
                # Common "not found" warning message.
                "warning: Skip unavailable {} manager.".format(mid) in result.stderr,
                # Common "not implemented" optional command warning message.
                bool(
                    re.search(
                        r"warning: (Sync|Cleanup) not implemented for {}.".format(mid),
                        result.stderr,
                    )
                ),
                # Stats line at the end of output.
                "{}: ".format(mid) in result.stdout.splitlines()[-1]
                if result.stdout
                else "",
                # Match output of managers command.
                bool(
                    re.search(
                        r"\s+│\s+{}\s+│\s+(✓|✘).+│\s+(✓|✘)".format(mid),
                        strip_ansi(result.stdout),
                    )
                ),
                # Sync command.
                "Sync {} package info...".format(mid) in result.stderr,
                # Upgrade command.
                "Updating all outdated packages from {}...".format(mid)
                in result.stderr,
                # Cleanup command.
                "Cleanup {}...".format(mid) in result.stderr,
                # Log message for backup command.
                "Dumping packages from {}...".format(mid) in result.stderr,
                # Restoring message.
                "Restore {} packages...".format(mid) in result.stderr,
                # Warning message for restore command.
                "warning: No [{}] section found.".format(mid) in result.stderr,
            ]

            if True in signals:
                found_managers.add(mid)
            else:
                skipped_managers.add(mid)

        # Compare managers reported by the CLI and those expected.
        assert found_managers == selected
        assert skipped_managers == MANAGER_IDS - selected

    @pytest.mark.parametrize("selector", ["--manager", "--exclude"])
    def test_invalid_manager_selector(self, invoke, subcmd, selector):
        result = invoke(selector, "unknown", subcmd)
        assert result.exit_code == 2
        assert not result.stdout
        assert "Error: Invalid value for " in result.stderr
        assert selector in result.stderr

    def test_default_all_managers(self, invoke, subcmd):
        """ Test all available managers are selected by default. """
        result = invoke(subcmd)
        assert result.exit_code == 0
        self.check_manager_selection(result)

    @pytest.mark.parametrize(
        "args,expected",
        [
            pytest.param(("--manager", "apm"), {"apm"}, id="single_selector"),
            pytest.param(("--manager", "apm") * 2, {"apm"}, id="duplicate_selectors"),
            pytest.param(
                ("--manager", "apm", "--manager", "gem"),
                {"apm", "gem"},
                id="multiple_selectors",
            ),
            pytest.param(
                ("--exclude", "apm"), MANAGER_IDS - {"apm"}, id="single_exclusion"
            ),
            pytest.param(
                ("--exclude", "apm") * 2,
                MANAGER_IDS - {"apm"},
                id="duplicate_exclusions",
            ),
            pytest.param(
                ("--exclude", "apm", "--exclude", "gem"),
                MANAGER_IDS - {"apm", "gem"},
                id="multiple_exclusions",
            ),
            pytest.param(
                ("--manager", "apm", "--exclude", "gem"), {"apm"}, id="priority_ordered"
            ),
            pytest.param(
                ("--exclude", "gem", "--manager", "apm"),
                {"apm"},
                id="priority_reversed",
            ),
            pytest.param(
                ("--manager", "apm", "--exclude", "apm"),
                set(),
                id="exclusion_override_ordered",
            ),
            pytest.param(
                ("--exclude", "apm", "--manager", "apm"),
                set(),
                id="exclusion_override_reversed",
            ),
        ],
    )
    def test_manager_selection(self, invoke, subcmd, args, expected):
        result = invoke(*args, subcmd)
        assert result.exit_code == 0
        self.check_manager_selection(result, expected)


class CLITableTests:

    """Test subcommands whose output is a configurable table.

    A table output is also allowed to be rendered as JSON.
    """

    # List of all supported rendering modes IDs, and their expected output.
    expected_renderings = {
        "ascii": ("---+---", None),
        "csv": (",", None),
        "csv-tab": ("\t", None),
        "double": ("═══╬═══", None),
        "fancy_grid": ("═══╪═══", None),
        "github": ("---|---", None),
        "grid": ("===+===", "ascii"),
        "html": ("<table>", None),
        "jira": (" || ", None),
        "json": ('": {', None),
        "latex": ("\\hline", None),
        "latex_booktabs": ("\\toprule", None),
        "mediawiki": ('{| class="wikitable" ', "jira"),
        "moinmoin": (" ''' || ''' ", "jira"),
        "orgtbl": ("---+---", None),
        "pipe": ("---|:---", None),
        "plain": ("  ", None),
        "psql": ("---+---", None),
        "rst": ("===  ===", None),
        "simple": ("---  ---", None),
        "textile": (" |_. ", None),
        "tsv": ("\t", None),
        "vertical": ("***[ 1. row ]***", None),
    }

    def test_recognized_modes(self):
        """Check all rendering modes proposed by the table module are
        accounted for."""
        assert RENDERING_MODES == set(self.expected_renderings.keys())

    def test_default_table_rendering(self, invoke, subcmd):
        """ Check default rendering is fancy_grid. """
        result = invoke(subcmd)
        assert result.exit_code == 0

        expected = self.expected_renderings["fancy_grid"][0]

        # If no package found, check that no table gets rendered. Else, check
        # the selected mode is indeed rendered in <stdout>, so the CLI result
        # can be grep-ed.
        if result.stdout.startswith("0 package total ("):
            assert expected not in result.stdout
        else:
            assert expected in result.stdout

        assert expected not in result.stderr

    @pytest.mark.parametrize(
        "mode,expected,conflict",
        [(k, v[0], v[1]) for k, v in expected_renderings.items()],
    )
    def test_all_table_rendering(self, invoke, subcmd, mode, expected, conflict):
        """Check that from all rendering modes, only the selected one appears
        in <stdout> and only there. Any other mode are not expected neither in
        <stdout> or <stderr>.
        """
        result = invoke("--output-format", mode, subcmd)
        assert result.exit_code == 0

        # If no package found, check that no table gets rendered. Else, check
        # the selected mode is indeed rendered in <stdout>, so the CLI result
        # can be grep-ed.
        if result.stdout.startswith("0 package total ("):
            # CSV mode will match on comma.
            if mode != "csv":
                assert expected not in result.stdout
        else:
            assert expected in result.stdout

        # Collect all possible unique traces from all possible rendering modes.
        blacklist = {
            v[0]
            for v in self.expected_renderings.values()
            # Exclude obvious character sequences shorter than 3 characters to
            # eliminate false negative.
            if len(v[0]) > 2
        }
        # Remove overlapping edge-cases.
        if conflict:
            blacklist.remove(self.expected_renderings[conflict][0])

        for unexpected in blacklist:
            if unexpected != expected:
                # The unexpected trace is not the selected one, it should not
                # appears at all in stdout.
                assert unexpected not in result.stdout
            # Any expected output from all rendering modes must not appears in
            # <stderr>, including the selected one.
            assert unexpected not in result.stderr

    def test_json_debug_output(self, invoke, subcmd):
        """Output is expected to be parseable if read from <stdout> even in
        debug level as these messages are redirected to <stderr>.

        Also checks that JSON output format is not supported by all commands.
        """
        result = invoke("--output-format", "json", "--verbosity", "DEBUG", subcmd)
        assert result.exit_code == 0
        assert "debug:" in result.stderr
        json.loads(result.stdout)
        json.loads(result.output)
        with pytest.raises(json.decoder.JSONDecodeError):
            json.loads(result.stderr)



# -*- coding: utf-8 -*-
#
# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
# All Rights Reserved.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import re

from ..base import PackageManager
from ..platform import LINUX
from ..version import TokenizedString, parse_version


class OPKG(PackageManager):

    platforms = frozenset([LINUX])

    requirement = "0.2.0"

    def get_version(self):
        """Fetch version from ``opkg --version`` output.

        Raw CLI output samples:

        .. code-block:: shell-session

            ► opkg --version
            opkg version 0.3.6 (libsolv 0.7.5)
        """
        output = self.run_cli("--version")
        if output:
            return parse_version(output.splitlines()[0].split()[2])

    def sync(self):
        """

        Raw CLI output samples:

        .. code-block:: shell-session
        """
        super(OPKG, self).sync()
        self.run_cli(self.global_args, "update")

    @property
    def installed(self):
        """Fetch installed packages from ``opkg list-installed`` output.

        Raw CLI output samples:

        .. code-block:: shell-session
            ► opkg list-installed
            3rd-party-feed-configs - 1.1-r0
            aio-grab - 1.0+git71+c79e264-r0
            alsa-conf - 1.1.9-r0
            alsa-state - 0.2.0-r5
            alsa-states - 0.2.0-r5
            alsa-utils-alsactl - 1.1.9-r0
            avahi-daemon - 0.7-r0
            base-files - 3.0.14-r89
            base-files-dev - 3.0.14-r89
            base-passwd - 3.5.29-r0
            bash - 5.0-r0
            bash-completion - 2.9-r0
            bash-completion-dev - 2.9-r0
            bash-dev - 5.0-r0
            binutils - 2.32.0-r0
            busybox - 1.31.0-r0
            busybox-inetd - 1.31.0-r0
            busybox-mdev - 1.31.0-r0
            busybox-syslog - 1.31.0-r0
            busybox-udhcpc - 1.31.0-r0
        """
        installed = {}

        output = self.run_cli(self.global_args, "list-installed")

        if output:
            regexp = re.compile(r"(\S+) - (\S+)")
            for package in output.splitlines():
                match = regexp.match(package)
                if match:
                    package_id, installed_version = match.groups()
                    installed[package_id] = {
                        "id": package_id,
                        "name": package_id,
                        "installed_version": parse_version(installed_version),
                    }

        return installed

    def search(self, query, extended, exact):
        """ Simulate search by listing all packages. """
        matches = {}

        # opkg doesn't have a working 'search', so get all packages and
        # filter the packages later.
        output = self.run_cli(self.global_args, "list")

        if output:
            regexp = re.compile(
                r"""
                (?P<package_id>\S+)
                \ -\
                (?P<version>\S+)
                \ -\
                (?P<description>.+)
                """,
                re.VERBOSE | re.MULTILINE,
            )

            for package_id, version, description in regexp.findall(output):

                # Skip all non-stricly matching package IDs in exact mode.
                if exact:
                    if query != package_id:
                        continue

                else:
                    # All other modes search for matching in package IDs and
                    # names.
                    searched_content = set(map(str, TokenizedString(package_id)))

                    # Also search within the description in extended mode.
                    if extended:
                        searched_content.update(
                            set(map(str, TokenizedString(description)))
                        )

                    # Skip package if not all sub-strings are present in the
                    # searched content.
                    query_parts = set(map(str, TokenizedString(query)))
                    if not query_parts.issubset(searched_content):
                        continue

                matches[package_id] = {
                    "id": package_id,
                    "name": package_id,
                    "latest_version": parse_version(version),
                }

        return matches

    @property
    def outdated(self):
        """Fetch outdated packages from ``opkg list-upgradable`` output.

        Raw CLI output samples:

        .. code-block:: shell-session

            ► opkg list-upgradable
            openpli-bootlogo - 20190717-r0 - 20190718-r0
            enigma2-hotplug - 2.7+git1720+55c6b34-r0 - 2.7+git1722+daf2f52-r0
        """
        outdated = {}

        output = self.run_cli(self.global_args, "list-upgradable")

        if output:
            regexp = re.compile(r"(\S+) - (\S+) - (\S+)")
            for package in output.splitlines():
                match = regexp.match(package)
                if match:
                    package_id, installed_version, latest_version = match.groups()
                    outdated[package_id] = {
                        "id": package_id,
                        "name": package_id,
                        "latest_version": parse_version(latest_version),
                        "installed_version": parse_version(installed_version),
                    }

        return outdated

    def upgrade_cli(self, package_id=None):
        cmd = [self.cli_path, self.global_args, "upgrade"]
        if package_id:
            cmd.append(package_id)
        return cmd

    def upgrade_all_cli(self):
        return self.upgrade_cli()
